"""
Tests for L{eliot.tai64n}.
"""

from __future__ import unicode_literals

import errno
import time
import subprocess
from unittest import TestCase, SkipTest

from ..tai64n import encode, decode


class CodecTests(TestCase):
    """
    Tests for L{encode} and L{decode}.
    """

    def test_encode(self):
        """
        L{encode} encodes timestamps in TAI64N format.
        """
        t = 1387299889.153187625
        self.assertEqual(encode(t), "@4000000052b0843b092174b9")

    def test_decode(self):
        """
        L{decode} decodes timestamps from TAI64N format.
        """
        t = time.time()
        self.assertAlmostEqual(t, decode(encode(t)), 9)


class FunctionalTests(TestCase):
    """
    Functional tests for L{encode}.
    """

    def test_encode(self):
        """
        The daemontools tai64nlocal tool can correctly decode timestamps output
        by L{encode}.
        """
        try:
            process = subprocess.Popen(
                ["tai64nlocal"],
                bufsize=4096,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise SkipTest("This test requires the daemontools package")
            else:
                raise
        # Because of limitations of the time libraries tai64nlocal uses we
        # apparently can't verify beyond this level of accuracy.
        timestamp = int(time.time()) + 0.12345
        process.stdin.write((encode(timestamp) + "\n").encode("ascii"))
        process.stdin.close()
        decodedToLocalTime = process.stdout.read().strip()
        self.assertEqual(
            time.strftime("%Y-%m-%d %H:%M:%S.12345", time.localtime(timestamp)).encode(
                "ascii"
            ),
            decodedToLocalTime[:25],
        )



"""Tests for standard library logging integration."""

from unittest import TestCase
import logging
import traceback

from ..testing import assertContainsFields, capture_logging
from ..stdlib import EliotHandler
from .test_traceback import assert_expected_traceback


class StdlibTests(TestCase):
    """Tests for stdlib integration."""

    @capture_logging(None)
    def test_handler(self, logger):
        """The EliotHandler routes messages to Eliot."""
        stdlib_logger = logging.getLogger("eliot-test")
        stdlib_logger.setLevel(logging.DEBUG)
        handler = EliotHandler()
        stdlib_logger.addHandler(handler)
        stdlib_logger.info("hello")
        stdlib_logger.warning("ono")
        message = logger.messages[0]
        assertContainsFields(
            self,
            message,
            {
                "message_type": "eliot:stdlib",
                "log_level": "INFO",
                "message": "hello",
                "logger": "eliot-test",
            },
        )
        message = logger.messages[1]
        assertContainsFields(
            self,
            message,
            {
                "message_type": "eliot:stdlib",
                "log_level": "WARNING",
                "message": "ono",
                "logger": "eliot-test",
            },
        )

    @capture_logging(None)
    def test_traceback(self, logger):
        """The EliotHandler routes tracebacks to Eliot."""
        stdlib_logger = logging.getLogger("eliot-test2")
        stdlib_logger.setLevel(logging.DEBUG)
        handler = EliotHandler()
        stdlib_logger.addHandler(handler)
        try:
            raise RuntimeError()
        except Exception as e:
            exception = e
            expected_traceback = traceback.format_exc()
            stdlib_logger.exception("ono")
        message = logger.messages[0]
        assertContainsFields(
            self,
            message,
            {
                "message_type": "eliot:stdlib",
                "log_level": "ERROR",
                "message": "ono",
                "logger": "eliot-test2",
            },
        )
        assert_expected_traceback(
            self, logger, logger.messages[1], exception, expected_traceback
        )



"""
Tests for L{eliot.twisted}.
"""

from __future__ import absolute_import, unicode_literals, print_function

import sys
from functools import wraps

try:
    from twisted.internet.defer import Deferred, succeed, fail, returnValue
    from twisted.trial.unittest import TestCase
    from twisted.python.failure import Failure
    from twisted.logger import globalLogPublisher
except ImportError:
    # Make tests not run at all.
    TestCase = object
else:
    # Make sure we always import this if Twisted is available, so broken
    # logwriter.py causes a failure:
    from ..twisted import (
        DeferredContext,
        AlreadyFinished,
        _passthrough,
        redirectLogsForTrial,
        _RedirectLogsForTrial,
        TwistedDestination,
        inline_callbacks,
    )

from .test_generators import assert_expected_action_tree

from .._action import start_action, current_action, Action, TaskLevel
from .._output import MemoryLogger, Logger
from .._message import Message
from ..testing import assertContainsFields, capture_logging
from .. import removeDestination, addDestination
from .._traceback import write_traceback
from .common import FakeSys


class PassthroughTests(TestCase):
    """
    Tests for L{_passthrough}.
    """

    def test_passthrough(self):
        """
        L{_passthrough} returns the passed-in value.
        """
        obj = object()
        self.assertIs(obj, _passthrough(obj))


def withActionContext(f):
    """
    Decorator that calls a function with an action context.

    @param f: A function.
    """
    logger = MemoryLogger()
    action = start_action(logger, "test")

    @wraps(f)
    def test(self):
        with action.context():
            return f(self)

    return test


class DeferredContextTests(TestCase):
    """
    Tests for L{DeferredContext}.
    """

    def test_requireContext(self):
        """
        L{DeferredContext} raises a L{RuntimeError} if it is called without an
        action context.
        """
        self.assertRaises(RuntimeError, DeferredContext, Deferred())

    @withActionContext
    def test_result(self):
        """
        The passed-in L{Deferred} is available as the L{DeferredContext}'s
        C{result} attribute.
        """
        result = Deferred()
        context = DeferredContext(result)
        self.assertIs(context.result, result)

    @withActionContext
    def test_addCallbacksCallbackToDeferred(self):
        """
        L{DeferredContext.addCallbacks} passes the given callback and its
        corresponding arguments to the wrapped L{Deferred}'s
        C{addCallbacks}.
        """
        called = []

        def f(value, x, y):
            called.append((value, x, y))

        result = Deferred()
        context = DeferredContext(result)
        context.addCallbacks(f, lambda x: None, (1,), {"y": 2})
        result.callback(0)
        self.assertEqual(called, [(0, 1, 2)])

    @withActionContext
    def test_addCallbacksErrbackToDeferred(self):
        """
        L{DeferredContext.addCallbacks} passes the given errback and its
        corresponding arguments to the wrapped L{Deferred}'s
        C{addCallbacks}.
        """
        called = []

        def f(value, x, y):
            value.trap(RuntimeError)
            called.append((x, y))

        result = Deferred()
        context = DeferredContext(result)
        context.addCallbacks(lambda x: None, f, None, None, (1,), {"y": 2})
        result.errback(RuntimeError())
        self.assertEqual(called, [(1, 2)])

    @withActionContext
    def test_addCallbacksWithOnlyCallback(self):
        """
        L{DeferredContext.addCallbacks} can be called with a single argument, a
        callback function, and passes it to the wrapped L{Deferred}'s
        C{addCallbacks}.
        """
        called = []

        def f(value):
            called.append(value)

        result = Deferred()
        context = DeferredContext(result)
        context.addCallbacks(f)
        result.callback(0)
        self.assertEqual(called, [0])

    @withActionContext
    def test_addCallbacksWithOnlyCallbackErrorCase(self):
        """
        L{DeferredContext.addCallbacks} can be called with a single argument, a
        callback function, and passes a pass-through errback to the wrapped
        L{Deferred}'s C{addCallbacks}.
        """
        called = []

        def f(value):
            called.append(value)

        class ExpectedException(Exception):
            pass

        result = Deferred()
        context = DeferredContext(result)
        context.addCallbacks(f)
        result.errback(Failure(ExpectedException()))
        self.assertEqual(called, [])
        # The assertion is inside `failureResultOf`.
        self.failureResultOf(result, ExpectedException)

    @withActionContext
    def test_addCallbacksReturnSelf(self):
        """
        L{DeferredContext.addCallbacks} returns the L{DeferredContext}.
        """
        result = Deferred()
        context = DeferredContext(result)
        self.assertIs(context, context.addCallbacks(lambda x: None, lambda x: None))

    def test_addCallbacksCallbackContext(self):
        """
        L{DeferedContext.addCallbacks} adds a callback that runs in context of
        action that the L{DeferredContext} was created with.
        """
        logger = MemoryLogger()
        action1 = start_action(logger, "test")
        action2 = start_action(logger, "test")
        context = []
        d = succeed(None)
        with action1.context():
            d = DeferredContext(d)
            with action2.context():
                d.addCallbacks(lambda x: context.append(current_action()), lambda x: x)
        self.assertEqual(context, [action1])

    def test_addCallbacksErrbackContext(self):
        """
        L{DeferedContext.addCallbacks} adds an errback that runs in context of
        action that the L{DeferredContext} was created with.
        """
        logger = MemoryLogger()
        action1 = start_action(logger, "test")
        action2 = start_action(logger, "test")
        context = []
        d = fail(RuntimeError())
        with action1.context():
            d = DeferredContext(d)
            with action2.context():
                d.addCallbacks(lambda x: x, lambda x: context.append(current_action()))
        self.assertEqual(context, [action1])

    @withActionContext
    def test_addCallbacksCallbackResult(self):
        """
        A callback added with DeferredContext.addCallbacks has its result
        passed on to the next callback.
        """
        d = succeed(0)
        d = DeferredContext(d)
        d.addCallbacks(lambda x: [x, 1], lambda x: x)
        self.assertEqual(self.successResultOf(d.result), [0, 1])

    @withActionContext
    def test_addCallbacksErrbackResult(self):
        """
        An errback added with DeferredContext.addCallbacks has its result
        passed on to the next callback.
        """
        exception = ZeroDivisionError()
        d = fail(exception)
        d = DeferredContext(d)
        d.addCallbacks(lambda x: x, lambda x: [x.value, 1])
        self.assertEqual(self.successResultOf(d.result), [exception, 1])

    def test_addActionFinishNoImmediateLogging(self):
        """
        L{DeferredContext.addActionFinish} does not log anything if the
        L{Deferred} hasn't fired yet.
        """
        d = Deferred()
        logger = MemoryLogger()
        action = Action(logger, "uuid", TaskLevel(level=[1]), "sys:me")
        with action.context():
            DeferredContext(d).addActionFinish()
        self.assertFalse(logger.messages)

    def test_addActionFinishSuccess(self):
        """
        When the L{Deferred} referred to by L{DeferredContext.addActionFinish}
        fires successfully, a finish message is logged.
        """
        d = Deferred()
        logger = MemoryLogger()
        action = Action(logger, "uuid", TaskLevel(level=[1]), "sys:me")
        with action.context():
            DeferredContext(d).addActionFinish()
        d.callback("result")
        assertContainsFields(
            self,
            logger.messages[0],
            {
                "task_uuid": "uuid",
                "task_level": [1, 1],
                "action_type": "sys:me",
                "action_status": "succeeded",
            },
        )

    def test_addActionFinishSuccessPassThrough(self):
        """
        L{DeferredContext.addActionFinish} passes through a successful result
        unchanged.
        """
        d = Deferred()
        logger = MemoryLogger()
        action = Action(logger, "uuid", TaskLevel(level=[1]), "sys:me")
        with action.context():
            DeferredContext(d).addActionFinish()
        d.callback("result")
        result = []
        d.addCallback(result.append)
        self.assertEqual(result, ["result"])

    def test_addActionFinishFailure(self):
        """
        When the L{Deferred} referred to in L{DeferredContext.addActionFinish}
        fires with an exception, a finish message is logged.
        """
        d = Deferred()
        logger = MemoryLogger()
        action = Action(logger, "uuid", TaskLevel(level=[1]), "sys:me")
        with action.context():
            DeferredContext(d).addActionFinish()
        exception = RuntimeError("because")
        d.errback(exception)
        assertContainsFields(
            self,
            logger.messages[0],
            {
                "task_uuid": "uuid",
                "task_level": [1, 1],
                "action_type": "sys:me",
                "action_status": "failed",
                "reason": "because",
                "exception": "%s.RuntimeError" % (RuntimeError.__module__,),
            },
        )
        d.addErrback(lambda _: None)  # don't let Failure go to Twisted logs

    def test_addActionFinishFailurePassThrough(self):
        """
        L{DeferredContext.addActionFinish} passes through a failed result
        unchanged.
        """
        d = Deferred()
        logger = MemoryLogger()
        action = Action(logger, "uuid", TaskLevel(level=[1]), "sys:me")
        with action.context():
            DeferredContext(d).addActionFinish()
        failure = Failure(RuntimeError())
        d.errback(failure)
        result = []
        d.addErrback(result.append)
        self.assertEqual(result, [failure])

    @withActionContext
    def test_addActionFinishRaisesAfterAddActionFinish(self):
        """
        After L{DeferredContext.addActionFinish} is called, additional calls to
        L{DeferredContext.addActionFinish} result in a L{AlreadyFinished}
        exception.
        """
        d = DeferredContext(Deferred())
        d.addActionFinish()
        self.assertRaises(AlreadyFinished, d.addActionFinish)

    @withActionContext
    def test_addCallbacksRaisesAfterAddActionFinish(self):
        """
        After L{DeferredContext.addActionFinish} is called, additional calls to
        L{DeferredContext.addCallbacks} result in a L{AlreadyFinished}
        exception.
        """
        d = DeferredContext(Deferred())
        d.addActionFinish()
        self.assertRaises(AlreadyFinished, d.addCallbacks, lambda x: x, lambda x: x)

    @withActionContext
    def test_addActionFinishResult(self):
        """
        L{DeferredContext.addActionFinish} returns the L{Deferred}.
        """
        d = Deferred()
        self.assertIs(d, DeferredContext(d).addActionFinish())

    # Having made sure DeferredContext.addCallbacks does the right thing
    # regarding action contexts, for addCallback/addErrback/addBoth we only
    # need to ensure that they call DeferredContext.addCallbacks.

    @withActionContext
    def test_addCallbackCallsAddCallbacks(self):
        """
        L{DeferredContext.addCallback} passes its arguments on to
        L{DeferredContext.addCallbacks}.
        """
        result = Deferred()
        context = DeferredContext(result)
        called = []

        def addCallbacks(
            callback,
            errback,
            callbackArgs=None,
            callbackKeywords=None,
            errbackArgs=None,
            errbackKeywords=None,
        ):
            called.append(
                (
                    callback,
                    errback,
                    callbackArgs,
                    callbackKeywords,
                    errbackArgs,
                    errbackKeywords,
                )
            )

        context.addCallbacks = addCallbacks

        def f(x, y, z):
            return None

        context.addCallback(f, 2, z=3)
        self.assertEqual(called, [(f, _passthrough, (2,), {"z": 3}, None, None)])

    @withActionContext
    def test_addCallbackReturnsSelf(self):
        """
        L{DeferredContext.addCallback} returns the L{DeferredContext}.
        """
        result = Deferred()
        context = DeferredContext(result)
        self.assertIs(context, context.addCallback(lambda x: None))

    @withActionContext
    def test_addErrbackCallsAddCallbacks(self):
        """
        L{DeferredContext.addErrback} passes its arguments on to
        L{DeferredContext.addCallbacks}.
        """
        result = Deferred()
        context = DeferredContext(result)
        called = []

        def addCallbacks(
            callback,
            errback,
            callbackArgs=None,
            callbackKeywords=None,
            errbackArgs=None,
            errbackKeywords=None,
        ):
            called.append(
                (
                    callback,
                    errback,
                    callbackArgs,
                    callbackKeywords,
                    errbackArgs,
                    errbackKeywords,
                )
            )

        context.addCallbacks = addCallbacks

        def f(x, y, z):
            pass

        context.addErrback(f, 2, z=3)
        self.assertEqual(called, [(_passthrough, f, None, None, (2,), {"z": 3})])

    @withActionContext
    def test_addErrbackReturnsSelf(self):
        """
        L{DeferredContext.addErrback} returns the L{DeferredContext}.
        """
        result = Deferred()
        context = DeferredContext(result)
        self.assertIs(context, context.addErrback(lambda x: None))

    @withActionContext
    def test_addBothCallsAddCallbacks(self):
        """
        L{DeferredContext.addBoth} passes its arguments on to
        L{DeferredContext.addCallbacks}.
        """
        result = Deferred()
        context = DeferredContext(result)
        called = []

        def addCallbacks(
            callback,
            errback,
            callbackArgs=None,
            callbackKeywords=None,
            errbackArgs=None,
            errbackKeywords=None,
        ):
            called.append(
                (
                    callback,
                    errback,
                    callbackArgs,
                    callbackKeywords,
                    errbackArgs,
                    errbackKeywords,
                )
            )

        context.addCallbacks = addCallbacks

        def f(x, y, z):
            return None

        context.addBoth(f, 2, z=3)
        self.assertEqual(called, [(f, f, (2,), {"z": 3}, (2,), {"z": 3})])

    @withActionContext
    def test_addBothReturnsSelf(self):
        """
        L{DeferredContext.addBoth} returns the L{DeferredContext}.
        """
        result = Deferred()
        context = DeferredContext(result)
        self.assertIs(context, context.addBoth(lambda x: None))


class RedirectLogsForTrialTests(TestCase):
    """
    Tests for L{redirectLogsForTrial}.
    """

    def assertDestinationAdded(self, programPath):
        """
        Assert that when running under the given program a new destination is
        added by L{redirectLogsForTrial}.

        @param programPath: A path to a program.
        @type programPath: L{str}
        """
        destination = _RedirectLogsForTrial(FakeSys([programPath], b""))()
        self.assertIsInstance(destination, TwistedDestination)
        # If this was not added as destination, removing it will raise an
        # exception:
        try:
            removeDestination(destination)
        except ValueError:
            self.fail("Destination was not added.")

    def test_withTrial(self):
        """
        When C{sys.argv[0]} is C{"trial"} a new destination is added by
        L{redirectLogsForTrial}.
        """
        self.assertDestinationAdded("trial")

    def test_withAbsoluteTrialPath(self):
        """
        When C{sys.argv[0]} is an absolute path ending with C{"trial"} a new
        destination is added by L{redirectLogsForTrial}.
        """
        self.assertDestinationAdded("/usr/bin/trial")

    def test_withRelativeTrialPath(self):
        """
        When C{sys.argv[0]} is a relative path ending with C{"trial"} a new
        destination is added by L{redirectLogsForTrial}.
        """
        self.assertDestinationAdded("./trial")

    def test_withoutTrialNoDestination(self):
        """
        When C{sys.argv[0]} is not C{"trial"} no destination is added by
        L{redirectLogsForTrial}.
        """
        originalDestinations = Logger._destinations._destinations[:]
        _RedirectLogsForTrial(FakeSys(["myprogram.py"], b""))()
        self.assertEqual(Logger._destinations._destinations, originalDestinations)

    def test_trialAsPathNoDestination(self):
        """
        When C{sys.argv[0]} has C{"trial"} as directory name but not program
        name no destination is added by L{redirectLogsForTrial}.
        """
        originalDestinations = Logger._destinations._destinations[:]
        _RedirectLogsForTrial(FakeSys(["./trial/myprogram.py"], b""))()
        self.assertEqual(Logger._destinations._destinations, originalDestinations)

    def test_withoutTrialResult(self):
        """
        When not running under I{trial} L{None} is returned.
        """
        self.assertIs(None, _RedirectLogsForTrial(FakeSys(["myprogram.py"], b""))())

    def test_noDuplicateAdds(self):
        """
        If a destination has already been added, calling
        L{redirectLogsForTrial} a second time does not add another destination.
        """
        redirect = _RedirectLogsForTrial(FakeSys(["trial"], b""))
        destination = redirect()
        self.addCleanup(removeDestination, destination)
        originalDestinations = Logger._destinations._destinations[:]
        redirect()
        self.assertEqual(Logger._destinations._destinations, originalDestinations)

    def test_noDuplicateAddsResult(self):
        """
        If a destination has already been added, calling
        L{redirectLogsForTrial} a second time returns L{None}.
        """
        redirect = _RedirectLogsForTrial(FakeSys(["trial"], b""))
        destination = redirect()
        self.addCleanup(removeDestination, destination)
        result = redirect()
        self.assertIs(result, None)

    def test_publicAPI(self):
        """
        L{redirectLogsForTrial} is an instance of L{_RedirectLogsForTrial}.
        """
        self.assertIsInstance(redirectLogsForTrial, _RedirectLogsForTrial)

    def test_defaults(self):
        """
        By default L{redirectLogsForTrial} looks at L{sys.argv}.
        """
        self.assertEqual(redirectLogsForTrial._sys, sys)


class TwistedDestinationTests(TestCase):
    """
    Tests for L{TwistedDestination}.
    """

    def redirect_to_twisted(self):
        """
        Redirect Eliot logs to Twisted.

        @return: L{list} of L{dict} - the log messages written to Twisted will
             eventually be appended to this list.
        """
        written = []

        def got_event(event):
            if event.get("log_namespace") == "eliot":
                written.append((event["log_level"].name, event["eliot"]))

        globalLogPublisher.addObserver(got_event)
        self.addCleanup(globalLogPublisher.removeObserver, got_event)
        destination = TwistedDestination()
        addDestination(destination)
        self.addCleanup(removeDestination, destination)
        return written

    def redirect_to_list(self):
        """
        Redirect Eliot logs to a list.

        @return: L{list} that will have eventually have the written Eliot
            messages added to it.
        """
        written = []
        destination = written.append
        addDestination(destination)
        self.addCleanup(removeDestination, destination)
        return written

    def test_normalMessages(self):
        """
        Regular eliot messages are pretty-printed to the given L{LogPublisher}.
        """
        writtenToTwisted = self.redirect_to_twisted()
        written = self.redirect_to_list()
        logger = Logger()
        Message.new(x=123, y=456).write(logger)
        self.assertEqual(writtenToTwisted, [("info", written[0])])

    def test_tracebackMessages(self):
        """
        Traceback eliot messages are written to the given L{LogPublisher} with
        the traceback formatted for easier reading.
        """
        writtenToTwisted = self.redirect_to_twisted()
        written = self.redirect_to_list()
        logger = Logger()

        def raiser():
            raise RuntimeError("because")

        try:
            raiser()
        except Exception:
            write_traceback(logger)
        self.assertEqual(writtenToTwisted, [("critical", written[0])])


class InlineCallbacksTests(TestCase):
    """Tests for C{inline_callbacks}."""

    # Get our custom assertion failure messages *and* the standard ones.
    longMessage = True

    def _a_b_test(self, logger, g):
        """A yield was done in between messages a and b inside C{inline_callbacks}."""
        with start_action(action_type="the-action"):
            self.assertIs(None, self.successResultOf(g()))
        assert_expected_action_tree(self, logger, "the-action", ["a", "yielded", "b"])

    @capture_logging(None)
    def test_yield_none(self, logger):
        def g():
            Message.log(message_type="a")
            yield
            Message.log(message_type="b")

        g = inline_callbacks(g, debug=True)

        self._a_b_test(logger, g)

    @capture_logging(None)
    def test_yield_fired_deferred(self, logger):
        def g():
            Message.log(message_type="a")
            yield succeed(None)
            Message.log(message_type="b")

        g = inline_callbacks(g, debug=True)

        self._a_b_test(logger, g)

    @capture_logging(None)
    def test_yield_unfired_deferred(self, logger):
        waiting = Deferred()

        def g():
            Message.log(message_type="a")
            yield waiting
            Message.log(message_type="b")

        g = inline_callbacks(g, debug=True)

        with start_action(action_type="the-action"):
            d = g()
            self.assertNoResult(waiting)
            waiting.callback(None)
            self.assertIs(None, self.successResultOf(d))
        assert_expected_action_tree(self, logger, "the-action", ["a", "yielded", "b"])

    @capture_logging(None)
    def test_returnValue(self, logger):
        result = object()

        @inline_callbacks
        def g():
            if False:
                yield
            returnValue(result)

        with start_action(action_type="the-action"):
            d = g()
            self.assertIs(result, self.successResultOf(d))

        assert_expected_action_tree(self, logger, "the-action", [])

    @capture_logging(None)
    def test_returnValue_in_action(self, logger):
        result = object()

        @inline_callbacks
        def g():
            if False:
                yield
            with start_action(action_type="g"):
                returnValue(result)

        with start_action(action_type="the-action"):
            d = g()
            self.assertIs(result, self.successResultOf(d))

        assert_expected_action_tree(self, logger, "the-action", [{"g": []}])

    @capture_logging(None)
    def test_nested_returnValue(self, logger):
        result = object()
        another = object()

        def g():
            d = h()
            # Run h through to the end but ignore its result.
            yield d
            # Give back _our_ result.
            returnValue(result)

        g = inline_callbacks(g, debug=True)

        def h():
            yield
            returnValue(another)

        h = inline_callbacks(h, debug=True)

        with start_action(action_type="the-action"):
            d = g()
            self.assertIs(result, self.successResultOf(d))

        assert_expected_action_tree(self, logger, "the-action", ["yielded", "yielded"])

    @capture_logging(None)
    def test_async_returnValue(self, logger):
        result = object()
        waiting = Deferred()

        @inline_callbacks
        def g():
            yield waiting
            returnValue(result)

        with start_action(action_type="the-action"):
            d = g()
            waiting.callback(None)
            self.assertIs(result, self.successResultOf(d))

    @capture_logging(None)
    def test_nested_async_returnValue(self, logger):
        result = object()
        another = object()

        waiting = Deferred()

        @inline_callbacks
        def g():
            yield h()
            returnValue(result)

        @inline_callbacks
        def h():
            yield waiting
            returnValue(another)

        with start_action(action_type="the-action"):
            d = g()
            waiting.callback(None)
            self.assertIs(result, self.successResultOf(d))



"""
Tests for L{eliot.testing}.
"""

from __future__ import unicode_literals

from unittest import SkipTest, TestResult, TestCase, skipUnless

try:
    import numpy as np
except ImportError:
    np = None

from ..testing import (
    issuperset,
    assertContainsFields,
    LoggedAction,
    LoggedMessage,
    validateLogging,
    UnflushedTracebacks,
    assertHasMessage,
    assertHasAction,
    validate_logging,
    capture_logging,
    swap_logger,
    check_for_errors,
)
from .._output import MemoryLogger
from .._action import start_action
from .._message import Message
from .._validation import ActionType, MessageType, ValidationError, Field
from .._traceback import write_traceback
from .. import add_destination, remove_destination, _output, log_message
from .common import CustomObject, CustomJSONEncoder


class IsSuperSetTests(TestCase):
    """
    Tests for L{issuperset}.
    """

    def test_equal(self):
        """
        Equal dictionaries are supersets of each other.
        """
        a = {"a": 1}
        b = a.copy()
        self.assertTrue(issuperset(a, b))

    def test_additionalIsSuperSet(self):
        """
        If C{A} is C{B} plus some extra entries, C{A} is superset of C{B}.
        """
        a = {"a": 1, "b": 2, "c": 3}
        b = {"a": 1, "c": 3}
        self.assertTrue(issuperset(a, b))

    def test_missingIsNotSuperSet(self):
        """
        If C{A} is C{B} minus some entries, C{A} is not a superset of C{B}.
        """
        a = {"a": 1, "c": 3}
        b = {"a": 1, "b": 2, "c": 3}
        self.assertFalse(issuperset(a, b))


class LoggedActionTests(TestCase):
    """
    Tests for L{LoggedAction}.
    """

    def test_values(self):
        """
        The values given to the L{LoggedAction} constructor are stored on it.
        """
        d1 = {"x": 1}
        d2 = {"y": 2}
        root = LoggedAction(d1, d2, [])
        self.assertEqual((root.startMessage, root.endMessage), (d1, d2))

    def fromMessagesIndex(self, messages, index):
        """
        Call L{LoggedAction.fromMessages} using action specified by index in
        a list of message dictionaries.

        @param messages: A C{list} of message dictionaries.

        @param index: Index to the logger's messages.

        @return: Result of L{LoggedAction.fromMessages}.
        """
        uuid = messages[index]["task_uuid"]
        level = messages[index]["task_level"]
        return LoggedAction.fromMessages(uuid, level, messages)

    def test_fromMessagesCreatesLoggedAction(self):
        """
        L{LoggedAction.fromMessages} returns a L{LoggedAction}.
        """
        logger = MemoryLogger()
        with start_action(logger, "test"):
            pass
        logged = self.fromMessagesIndex(logger.messages, 0)
        self.assertIsInstance(logged, LoggedAction)

    def test_fromMessagesStartAndSuccessfulFinish(self):
        """
        L{LoggedAction.fromMessages} finds the start and successful finish
        messages of an action and stores them in the result.
        """
        logger = MemoryLogger()
        Message.new(x=1).write(logger)
        with start_action(logger, "test"):
            Message.new(x=1).write(logger)
        # Now we should have x message, start action message, another x message
        # and finally finish message.
        logged = self.fromMessagesIndex(logger.messages, 1)
        self.assertEqual(
            (logged.startMessage, logged.endMessage),
            (logger.messages[1], logger.messages[3]),
        )

    def test_fromMessagesStartAndErrorFinish(self):
        """
        L{LoggedAction.fromMessages} finds the start and successful finish
        messages of an action and stores them in the result.
        """
        logger = MemoryLogger()
        try:
            with start_action(logger, "test"):
                raise KeyError()
        except KeyError:
            pass
        logged = self.fromMessagesIndex(logger.messages, 0)
        self.assertEqual(
            (logged.startMessage, logged.endMessage),
            (logger.messages[0], logger.messages[1]),
        )

    def test_fromMessagesStartNotFound(self):
        """
        L{LoggedAction.fromMessages} raises a L{ValueError} if a start message
        is not found.
        """
        logger = MemoryLogger()
        with start_action(logger, action_type="test"):
            pass
        self.assertRaises(ValueError, self.fromMessagesIndex, logger.messages[1:], 0)

    def test_fromMessagesFinishNotFound(self):
        """
        L{LoggedAction.fromMessages} raises a L{ValueError} if a finish message
        is not found.
        """
        logger = MemoryLogger()
        with start_action(logger, action_type="test"):
            pass
        with self.assertRaises(ValueError) as cm:
            self.fromMessagesIndex(logger.messages[:1], 0)
        self.assertEqual(cm.exception.args[0], "Missing end message of type test")

    def test_fromMessagesAddsChildMessages(self):
        """
        L{LoggedAction.fromMessages} adds direct child messages to the
        constructed L{LoggedAction}.
        """
        logger = MemoryLogger()
        # index 0:
        Message.new(x=1).write(logger)
        # index 1 - start action
        with start_action(logger, "test"):
            # index 2
            Message.new(x=2).write(logger)
            # index 3
            Message.new(x=3).write(logger)
        # index 4 - end action
        # index 5
        Message.new(x=4).write(logger)
        logged = self.fromMessagesIndex(logger.messages, 1)

        expectedChildren = [
            LoggedMessage(logger.messages[2]),
            LoggedMessage(logger.messages[3]),
        ]
        self.assertEqual(logged.children, expectedChildren)

    def test_fromMessagesAddsChildActions(self):
        """
        L{LoggedAction.fromMessages} recursively adds direct child actions to
        the constructed L{LoggedAction}.
        """
        logger = MemoryLogger()
        # index 0
        with start_action(logger, "test"):
            # index 1:
            with start_action(logger, "test2"):
                # index 2
                Message.new(message_type="end", x=2).write(logger)
            # index 3 - end action
            with start_action(logger, "test3"):
                # index 4
                pass
            # index 5 - end action
        # index 6 - end action
        logged = self.fromMessagesIndex(logger.messages, 0)

        self.assertEqual(logged.children[0], self.fromMessagesIndex(logger.messages, 1))
        self.assertEqual(
            logged.type_tree(), {"test": [{"test2": ["end"]}, {"test3": []}]}
        )

    def test_ofType(self):
        """
        L{LoggedAction.ofType} returns a list of L{LoggedAction} created by the
        specified L{ActionType}.
        """
        ACTION = ActionType("myaction", [], [], "An action!")
        logger = MemoryLogger()
        # index 0
        with start_action(logger, "test"):
            # index 1:
            with ACTION(logger):
                # index 2
                Message.new(x=2).write(logger)
            # index 3 - end action
            # index 4 - end action
            # index 5
        with ACTION(logger):
            pass
        # index 6 - end action
        logged = LoggedAction.ofType(logger.messages, ACTION)
        self.assertEqual(
            logged,
            [
                self.fromMessagesIndex(logger.messages, 1),
                self.fromMessagesIndex(logger.messages, 5),
            ],
        )

        # String-variant of ofType:
        logged2 = LoggedAction.ofType(logger.messages, "myaction")
        self.assertEqual(logged, logged2)

    def test_ofTypeNotFound(self):
        """
        L{LoggedAction.ofType} returns an empty list if actions of the given
        type cannot be found.
        """
        ACTION = ActionType("myaction", [], [], "An action!")
        logger = MemoryLogger()
        self.assertEqual(LoggedAction.ofType(logger.messages, ACTION), [])

    def test_descendants(self):
        """
        L{LoggedAction.descendants} returns all descendants of the
        L{LoggedAction}.
        """
        ACTION = ActionType("myaction", [], [], "An action!")
        logger = MemoryLogger()
        # index 0
        with ACTION(logger):
            # index 1:
            with start_action(logger, "test"):
                # index 2
                Message.new(x=2).write(logger)
            # index 3 - end action
            # index 4
            Message.new(x=2).write(logger)
        # index 5 - end action

        loggedAction = LoggedAction.ofType(logger.messages, ACTION)[0]
        self.assertEqual(
            list(loggedAction.descendants()),
            [
                self.fromMessagesIndex(logger.messages, 1),
                LoggedMessage(logger.messages[2]),
                LoggedMessage(logger.messages[4]),
            ],
        )

    def test_succeeded(self):
        """
        If the action succeeded, L{LoggedAction.succeeded} will be true.
        """
        logger = MemoryLogger()
        with start_action(logger, "test"):
            pass
        logged = self.fromMessagesIndex(logger.messages, 0)
        self.assertTrue(logged.succeeded)

    def test_notSucceeded(self):
        """
        If the action failed, L{LoggedAction.succeeded} will be false.
        """
        logger = MemoryLogger()
        try:
            with start_action(logger, "test"):
                raise KeyError()
        except KeyError:
            pass
        logged = self.fromMessagesIndex(logger.messages, 0)
        self.assertFalse(logged.succeeded)


class LoggedMessageTest(TestCase):
    """
    Tests for L{LoggedMessage}.
    """

    def test_values(self):
        """
        The values given to the L{LoggedMessage} constructor are stored on it.
        """
        message = {"x": 1}
        logged = LoggedMessage(message)
        self.assertEqual(logged.message, message)

    def test_ofType(self):
        """
        L{LoggedMessage.ofType} returns a list of L{LoggedMessage} created by the
        specified L{MessageType}.
        """
        MESSAGE = MessageType("mymessage", [], "A message!")
        logger = MemoryLogger()
        # index 0
        MESSAGE().write(logger)
        # index 1
        Message.new(x=2).write(logger)
        # index 2
        MESSAGE().write(logger)
        logged = LoggedMessage.ofType(logger.messages, MESSAGE)
        self.assertEqual(
            logged,
            [LoggedMessage(logger.messages[0]), LoggedMessage(logger.messages[2])],
        )

        # Lookup by string type:
        logged2 = LoggedMessage.ofType(logger.messages, "mymessage")
        self.assertEqual(logged, logged2)

    def test_ofTypeNotFound(self):
        """
        L{LoggedMessage.ofType} returns an empty list if messages of the given
        type cannot be found.
        """
        MESSAGE = MessageType("mymessage", [], "A message!")
        logger = MemoryLogger()
        self.assertEqual(LoggedMessage.ofType(logger.messages, MESSAGE), [])


class AssertContainsFields(TestCase):
    """
    Tests for L{assertContainsFields}.
    """

    class ContainsTest(TestCase):
        """
        A test case that uses L{assertContainsFields}.
        """

        def __init__(self, message, expectedFields):
            TestCase.__init__(self)
            self.message = message
            self.expectedFields = expectedFields

        def runTest(self):
            assertContainsFields(self, self.message, self.expectedFields)

    def test_equal(self):
        """
        Equal dictionaries contain each other.
        """
        message = {"a": 1}
        expected = message.copy()
        test = self.ContainsTest(message, expected)
        # No exception raised:
        test.debug()

    def test_additionalIsSuperSet(self):
        """
        If C{A} is C{B} plus some extra entries, C{A} contains the fields in
        C{B}.
        """
        message = {"a": 1, "b": 2, "c": 3}
        expected = {"a": 1, "c": 3}
        test = self.ContainsTest(message, expected)
        # No exception raised:
        test.debug()

    def test_missingFields(self):
        """
        If C{A} is C{B} minus some entries, C{A} does not contain the fields in
        C{B}.
        """
        message = {"a": 1, "c": 3}
        expected = {"a": 1, "b": 2, "c": 3}
        test = self.ContainsTest(message, expected)
        self.assertRaises(AssertionError, test.debug)

    def test_differentValues(self):
        """
        If C{A} has a different value for a specific field than C{B}, C{A} does
        not contain the fields in C{B}.
        """
        message = {"a": 1, "c": 3}
        expected = {"a": 1, "c": 2}
        test = self.ContainsTest(message, expected)
        self.assertRaises(AssertionError, test.debug)


class ValidateLoggingTestsMixin(object):
    """
    Tests for L{validateLogging} and L{capture_logging}.
    """

    validate = None

    def test_decoratedFunctionCalledWithMemoryLogger(self):
        """
        The underlying function decorated with L{validateLogging} is called with
        a L{MemoryLogger} instance.
        """
        result = []

        class MyTest(TestCase):
            @self.validate(None)
            def test_foo(this, logger):
                result.append((this, logger.__class__))

        theTest = MyTest("test_foo")
        theTest.run()
        self.assertEqual(result, [(theTest, MemoryLogger)])

    def test_decorated_function_passthrough(self):
        """
        Additional arguments are passed to the underlying function.
        """
        result = []

        def another_wrapper(f):
            def g(this):
                f(this, 1, 2, c=3)

            return g

        class MyTest(TestCase):
            @another_wrapper
            @self.validate(None)
            def test_foo(this, a, b, logger, c=None):
                result.append((a, b, c))

        theTest = MyTest("test_foo")
        theTest.debug()
        self.assertEqual(result, [(1, 2, 3)])

    def test_newMemoryLogger(self):
        """
        The underlying function decorated with L{validateLogging} is called with
        a new L{MemoryLogger} every time the wrapper is called.
        """
        result = []

        class MyTest(TestCase):
            @self.validate(None)
            def test_foo(this, logger):
                result.append(logger)

        theTest = MyTest("test_foo")
        theTest.run()
        theTest.run()
        self.assertIsNot(result[0], result[1])

    def test_returns(self):
        """
        The result of the underlying function is returned by wrapper when called.
        """

        class MyTest(TestCase):
            @self.validate(None)
            def test_foo(self, logger):
                return 123

        self.assertEqual(MyTest("test_foo").test_foo(), 123)

    def test_raises(self):
        """
        The exception raised by the underlying function is passed through by the
        wrapper when called.
        """
        exc = Exception()

        class MyTest(TestCase):
            @self.validate(None)
            def test_foo(self, logger):
                raise exc

        raised = None
        try:
            MyTest("test_foo").debug()
        except Exception as e:
            raised = e
        self.assertIs(exc, raised)

    def test_name(self):
        """
        The wrapper has the same name as the wrapped function.
        """

        class MyTest(TestCase):
            @self.validate(None)
            def test_foo(self, logger):
                pass

        self.assertEqual(MyTest.test_foo.__name__, "test_foo")

    def test_addCleanupValidate(self):
        """
        When a test method is decorated with L{validateLogging} it has
        L{MemoryLogger.validate} registered as a test cleanup.
        """
        MESSAGE = MessageType("mymessage", [], "A message")

        class MyTest(TestCase):
            @self.validate(None)
            def runTest(self, logger):
                self.logger = logger
                logger.write({"message_type": "wrongmessage"}, MESSAGE._serializer)

        test = MyTest()
        with self.assertRaises(ValidationError) as context:
            test.debug()
        # Some reference to the reason:
        self.assertIn("wrongmessage", str(context.exception))
        # Some reference to which file caused the problem:
        self.assertIn("test_testing.py", str(context.exception))

    def test_addCleanupTracebacks(self):
        """
        When a test method is decorated with L{validateLogging} it has has a
        check unflushed tracebacks in the L{MemoryLogger} registered as a
        test cleanup.
        """

        class MyTest(TestCase):
            @self.validate(None)
            def runTest(self, logger):
                try:
                    1 / 0
                except ZeroDivisionError:
                    write_traceback(logger)

        test = MyTest()
        self.assertRaises(UnflushedTracebacks, test.debug)

    def test_assertion(self):
        """
        If a callable is passed to L{validateLogging}, it is called with the
        L{TestCase} instance and the L{MemoryLogger} passed to the test
        method.
        """
        result = []

        class MyTest(TestCase):
            def assertLogging(self, logger):
                result.append((self, logger))

            @self.validate(assertLogging)
            def runTest(self, logger):
                self.logger = logger

        test = MyTest()
        test.run()
        self.assertEqual(result, [(test, test.logger)])

    def test_assertionArguments(self):
        """
        If a callable together with additional arguments and keyword arguments are
        passed to L{validateLogging}, the callable is called with the additional
        args and kwargs.
        """
        result = []

        class MyTest(TestCase):
            def assertLogging(self, logger, x, y):
                result.append((self, logger, x, y))

            @self.validate(assertLogging, 1, y=2)
            def runTest(self, logger):
                self.logger = logger

        test = MyTest()
        test.run()
        self.assertEqual(result, [(test, test.logger, 1, 2)])

    def test_assertionAfterTest(self):
        """
        If a callable is passed to L{validateLogging}, it is called with the
        after the main test code has run, allowing it to make assertions
        about log messages from the test.
        """

        class MyTest(TestCase):
            def assertLogging(self, logger):
                self.result.append(2)

            @self.validate(assertLogging)
            def runTest(self, logger):
                self.result = [1]

        test = MyTest()
        test.run()
        self.assertEqual(test.result, [1, 2])

    def test_assertionBeforeTracebackCleanup(self):
        """
        If a callable is passed to L{validateLogging}, it is called with the
        before the check for unflushed tracebacks, allowing it to flush
        traceback log messages.
        """

        class MyTest(TestCase):
            def assertLogging(self, logger):
                logger.flushTracebacks(ZeroDivisionError)
                self.flushed = True

            @self.validate(assertLogging)
            def runTest(self, logger):
                self.flushed = False
                try:
                    1 / 0
                except ZeroDivisionError:
                    write_traceback(logger)

        test = MyTest()
        test.run()
        self.assertTrue(test.flushed)


class ValidateLoggingTests(ValidateLoggingTestsMixin, TestCase):
    """
    Tests for L{validate_logging}.
    """

    validate = staticmethod(validate_logging)


class CaptureLoggingTests(ValidateLoggingTestsMixin, TestCase):
    """
    Tests for L{capture_logging}.
    """

    validate = staticmethod(capture_logging)

    def setUp(self):
        # Since we're not always calling the test method via the TestCase
        # infrastructure, sometimes cleanup methods are not called. This
        # means the original default logger is not restored. So we do so
        # manually. If the issue is a bug in capture_logging itself the
        # tests below will catch that.
        original_logger = _output._DEFAULT_LOGGER

        def cleanup():
            _output._DEFAULT_LOGGER = original_logger

        self.addCleanup(cleanup)

    def test_default_logger(self):
        """
        L{capture_logging} captures messages from logging that
        doesn't specify a L{Logger}.
        """

        class MyTest(TestCase):
            @capture_logging(None)
            def runTest(self, logger):
                Message.log(some_key=1234)
                self.logger = logger

        test = MyTest()
        test.run()
        self.assertEqual(test.logger.messages[0]["some_key"], 1234)

    def test_global_cleanup(self):
        """
        After the function wrapped with L{capture_logging} finishes,
        logging that doesn't specify a logger is logged normally.
        """

        class MyTest(TestCase):
            @capture_logging(None)
            def runTest(self, logger):
                pass

        test = MyTest()
        test.run()
        messages = []
        add_destination(messages.append)
        self.addCleanup(remove_destination, messages.append)
        Message.log(some_key=1234)
        self.assertEqual(messages[0]["some_key"], 1234)

    def test_global_cleanup_exception(self):
        """
        If the function wrapped with L{capture_logging} throws an exception,
        logging that doesn't specify a logger is logged normally.
        """

        class MyTest(TestCase):
            @capture_logging(None)
            def runTest(self, logger):
                raise RuntimeError()

        test = MyTest()
        test.run()
        messages = []
        add_destination(messages.append)
        self.addCleanup(remove_destination, messages.append)
        Message.log(some_key=1234)
        self.assertEqual(messages[0]["some_key"], 1234)

    def test_validationNotRunForSkip(self):
        """
        If the decorated test raises L{SkipTest} then the logging validation is
        also skipped.
        """

        class MyTest(TestCase):
            recorded = False

            def record(self, logger):
                self.recorded = True

            @validateLogging(record)
            def runTest(self, logger):
                raise SkipTest("Do not run this test.")

        test = MyTest()
        result = TestResult()
        test.run(result)

        # Verify that the validation function did not run and that the test was
        # nevertheless marked as a skip with the correct reason.
        self.assertEqual(
            (test.recorded, result.skipped, result.errors, result.failures),
            (False, [(test, "Do not run this test.")], [], []),
        )


class JSONEncodingTests(TestCase):
    """Tests for L{capture_logging} JSON encoder support."""

    @skipUnless(np, "NumPy is not installed.")
    @capture_logging(None)
    def test_default_JSON_encoder(self, logger):
        """
        L{capture_logging} validates using L{EliotJSONEncoder} by default.
        """
        # Default JSON encoder can't handle NumPy:
        log_message(message_type="hello", number=np.uint32(12))

    @capture_logging(None, encoder_=CustomJSONEncoder)
    def test_custom_JSON_encoder(self, logger):
        """
        L{capture_logging} can be called with a custom JSON encoder, which is then
        used for validation.
        """
        # Default JSON encoder can't handle this custom object:
        log_message(message_type="hello", object=CustomObject())


MESSAGE1 = MessageType(
    "message1", [Field.forTypes("x", [int], "A number")], "A message for testing."
)
MESSAGE2 = MessageType("message2", [], "A message for testing.")


class AssertHasMessageTests(TestCase):
    """
    Tests for L{assertHasMessage}.
    """

    class UnitTest(TestCase):
        """
        Test case that can be instantiated.
        """

        def runTest(self):
            pass

    def test_failIfNoMessagesOfType(self):
        """
        L{assertHasMessage} raises L{AssertionError} if the given L{MemoryLogger}
        has no messages of the given L{MessageType}.
        """
        test = self.UnitTest()
        logger = MemoryLogger()
        MESSAGE1(x=123).write(logger)
        self.assertRaises(AssertionError, assertHasMessage, test, logger, MESSAGE2)

    def test_returnsIfMessagesOfType(self):
        """
        L{assertHasMessage} returns the first message of the given L{MessageType}.
        """
        test = self.UnitTest()
        logger = MemoryLogger()
        MESSAGE1(x=123).write(logger)
        self.assertEqual(
            assertHasMessage(test, logger, MESSAGE1),
            LoggedMessage.ofType(logger.messages, MESSAGE1)[0],
        )

    def test_failIfNotSubset(self):
        """
        L{assertHasMessage} raises L{AssertionError} if the found message doesn't
        contain the given fields.
        """
        test = self.UnitTest()
        logger = MemoryLogger()
        MESSAGE1(x=123).write(logger)
        self.assertRaises(
            AssertionError, assertHasMessage, test, logger, MESSAGE1, {"x": 24}
        )

    def test_returnsIfSubset(self):
        """
        L{assertHasMessage} returns the first message of the given L{MessageType} if
        it contains the given fields.
        """
        test = self.UnitTest()
        logger = MemoryLogger()
        MESSAGE1(x=123).write(logger)
        self.assertEqual(
            assertHasMessage(test, logger, MESSAGE1, {"x": 123}),
            LoggedMessage.ofType(logger.messages, MESSAGE1)[0],
        )


ACTION1 = ActionType(
    "action1",
    [Field.forTypes("x", [int], "A number")],
    [Field.forTypes("result", [int], "A number")],
    "A action for testing.",
)
ACTION2 = ActionType("action2", [], [], "A action for testing.")


class AssertHasActionTests(TestCase):
    """
    Tests for L{assertHasAction}.
    """

    class UnitTest(TestCase):
        """
        Test case that can be instantiated.
        """

        def runTest(self):
            pass

    def test_failIfNoActionsOfType(self):
        """
        L{assertHasAction} raises L{AssertionError} if the given L{MemoryLogger}
        has no actions of the given L{ActionType}.
        """
        test = self.UnitTest()
        logger = MemoryLogger()
        with ACTION1(logger, x=123):
            pass
        self.assertRaises(AssertionError, assertHasAction, test, logger, ACTION2, True)

    def test_failIfWrongSuccessStatus(self):
        """
        L{assertHasAction} raises L{AssertionError} if the given success status does
        not match that of the found actions.
        """
        test = self.UnitTest()
        logger = MemoryLogger()
        with ACTION1(logger, x=123):
            pass
        try:
            with ACTION2(logger):
                1 / 0
        except ZeroDivisionError:
            pass
        self.assertRaises(AssertionError, assertHasAction, test, logger, ACTION1, False)
        self.assertRaises(AssertionError, assertHasAction, test, logger, ACTION2, True)

    def test_returnsIfMessagesOfType(self):
        """
        A successful L{assertHasAction} returns the first message of the given
        L{ActionType}.
        """
        test = self.UnitTest()
        logger = MemoryLogger()
        with ACTION1(logger, x=123):
            pass
        self.assertEqual(
            assertHasAction(test, logger, ACTION1, True),
            LoggedAction.ofType(logger.messages, ACTION1)[0],
        )

    def test_failIfNotStartSubset(self):
        """
        L{assertHasAction} raises L{AssertionError} if the found action doesn't
        contain the given start fields.
        """
        test = self.UnitTest()
        logger = MemoryLogger()
        with ACTION1(logger, x=123):
            pass
        self.assertRaises(
            AssertionError, assertHasAction, test, logger, ACTION1, True, {"x": 24}
        )

    def test_failIfNotEndSubset(self):
        """
        L{assertHasAction} raises L{AssertionError} if the found action doesn't
        contain the given end fields.
        """
        test = self.UnitTest()
        logger = MemoryLogger()
        with ACTION1(logger, x=123) as act:
            act.addSuccessFields(result=5)
        self.assertRaises(
            AssertionError,
            assertHasAction,
            test,
            logger,
            ACTION1,
            True,
            startFields={"x": 123},
            endFields={"result": 24},
        )

    def test_returns(self):
        """
        A successful L{assertHasAction} returns the first message of the given
        L{ActionType} after doing all validation.
        """
        test = self.UnitTest()
        logger = MemoryLogger()
        with ACTION1(logger, x=123) as act:
            act.addSuccessFields(result=5)
        self.assertEqual(
            assertHasAction(test, logger, ACTION1, True, {"x": 123}, {"result": 5}),
            LoggedAction.ofType(logger.messages, ACTION1)[0],
        )


class PEP8Tests(TestCase):
    """
    Tests for PEP 8 method compatibility.
    """

    def test_LoggedAction_from_messages(self):
        """
        L{LoggedAction.from_messages} is the same as
        L{LoggedAction.fromMessages}.
        """
        self.assertEqual(LoggedAction.from_messages, LoggedAction.fromMessages)

    def test_LoggedAction_of_type(self):
        """
        L{LoggedAction.of_type} is the same as
        L{LoggedAction.ofType}.
        """
        self.assertEqual(LoggedAction.of_type, LoggedAction.ofType)

    def test_LoggedAction_end_message(self):
        """
        L{LoggedAction.end_message} is the same as L{LoggedAction.endMessage}.
        """
        action = LoggedAction({1: 2}, {3: 4}, [])
        self.assertEqual(action.end_message, action.endMessage)

    def test_LoggedAction_start_message(self):
        """
        L{LoggedAction.start_message} is the same as
        L{LoggedAction.startMessage}.
        """
        action = LoggedAction({1: 2}, {3: 4}, [])
        self.assertEqual(action.start_message, action.startMessage)

    def test_LoggedMessage_of_type(self):
        """
        L{LoggedMessage.of_type} is the same as
        L{LoggedMessage.ofType}.
        """
        self.assertEqual(LoggedMessage.of_type, LoggedMessage.ofType)

    def test_validate_logging(self):
        """
        L{validate_logging} is the same as L{validateLogging}.
        """
        self.assertEqual(validate_logging, validateLogging)


class LowLevelTestingHooks(TestCase):
    """Tests for lower-level APIs for setting up MemoryLogger."""

    @capture_logging(None)
    def test_swap_logger(self, logger):
        """C{swap_logger} swaps out the current logger."""
        new_logger = MemoryLogger()
        old_logger = swap_logger(new_logger)
        Message.log(message_type="hello")

        # We swapped out old logger for new:
        self.assertIs(old_logger, logger)
        self.assertEqual(new_logger.messages[0]["message_type"], "hello")

        # Now restore old logger:
        intermediate_logger = swap_logger(old_logger)
        Message.log(message_type="goodbye")
        self.assertIs(intermediate_logger, new_logger)
        self.assertEqual(logger.messages[0]["message_type"], "goodbye")

    def test_check_for_errors_unflushed_tracebacks(self):
        """C{check_for_errors} raises on unflushed tracebacks."""
        logger = MemoryLogger()

        # No errors initially:
        check_for_errors(logger)

        try:
            1 / 0
        except ZeroDivisionError:
            write_traceback(logger)
        logger.flush_tracebacks(ZeroDivisionError)

        # Flushed tracebacks don't count:
        check_for_errors(logger)

        # But unflushed tracebacks do:
        try:
            raise RuntimeError
        except RuntimeError:
            write_traceback(logger)
        with self.assertRaises(UnflushedTracebacks):
            check_for_errors(logger)

    def test_check_for_errors_validation(self):
        """C{check_for_errors} raises on validation errors."""
        logger = MemoryLogger()
        logger.write({"x": 1, "message_type": "mem"})

        # No errors:
        check_for_errors(logger)

        # Now long something unserializable to JSON:
        logger.write({"message_type": object()})
        with self.assertRaises(TypeError):
            check_for_errors(logger)



"""
Command line program for filtering line-based Eliot logs.
"""

from __future__ import unicode_literals, absolute_import

if __name__ == "__main__":
    import eliot.filter

    eliot.filter.main()

import sys
from datetime import datetime, timedelta
from json import JSONEncoder

from ._bytesjson import dumps, loads


class _DatetimeJSONEncoder(JSONEncoder):
    """
    JSON encoder that supports L{datetime}.
    """

    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        return JSONEncoder.default(self, o)


class EliotFilter(object):
    """
    Filter Eliot log lines using a Python expression.

    @ivar code: A Python code object, the compiled filter expression.
    """

    _SKIP = object()

    def __init__(self, expr, incoming, output):
        """
        @param expr: A Python expression that will be called for each log message.
        @type expr: L{str}

        @param incoming: An iterable of L{bytes}, each of which is a serialized
            Eliot message.

        @param output: A file to which output should be written.
        @type output: L{file} or a file-like object.
        """
        self.code = compile(expr, "<string>", "eval")
        self.incoming = incoming
        self.output = output

    def run(self):
        """
        For each incoming message, decode the JSON, evaluate expression, encode
        as JSON and write that to the output file.
        """
        for line in self.incoming:
            message = loads(line)
            result = self._evaluate(message)
            if result is self._SKIP:
                continue
            self.output.write(dumps(result, cls=_DatetimeJSONEncoder) + b"\n")

    def _evaluate(self, message):
        """
        Evaluate the expression with the given Python object in its locals.

        @param message: A decoded JSON input.

        @return: The resulting object.
        """
        return eval(
            self.code,
            globals(),
            {
                "J": message,
                "timedelta": timedelta,
                "datetime": datetime,
                "SKIP": self._SKIP,
            },
        )


USAGE = b"""\
Usage: cat eliot.log | python -m eliot.filter <expr>

Read JSON-expression per line from stdin, and filter it using a Python
expression <expr>.

The expression will have a local `J` containing decoded JSON. `datetime` and
`timedelta` from Python's `datetime` module are also available as locals,
containing the corresponding classes. `SKIP` is also available, if it's the
expression result that indicates nothing should be output.

The output will be written to stdout using JSON serialization. `datetime`
objects will be serialized to ISO format.

Examples:

- Pass through the messages unchanged:

    $ cat eliot.log | python -m eliot.filter J

- Retrieve a specific field from a specific message type, dropping messages
  of other types:

    $ cat eliot.log | python -m eliot.filter \\
        "J['field'] if J.get('message_type') == 'my:message' else SKIP"
"""


def main(sys=sys):
    """
    Run the program.

    Accept arguments from L{sys.argv}, read from L{sys.stdin}, write to
    L{sys.stdout}.

    @param sys: An object with same interface and defaulting to the L{sys}
        module.
    """
    if len(sys.argv) != 2:
        sys.stderr.write(USAGE)
        return 1
    EliotFilter(sys.argv[1], sys.stdin, sys.stdout).run()
    return 0



"""
Tests for L{eliot._validation}.
"""

from __future__ import unicode_literals

from unittest import TestCase

from six import text_type as unicode

from .._validation import (
    Field,
    MessageType,
    ActionType,
    ValidationError,
    fields,
    _MessageSerializer,
)
from .._action import start_action, startTask
from .._output import MemoryLogger
from ..serializers import identity
from .. import add_destination, remove_destination


class TypedFieldTests(TestCase):
    """
    Tests for L{Field.forTypes}.
    """

    def test_validateCorrectType(self):
        """
        L{Field.validate} will not raise an exception if the given value is in
        the list of supported classes.
        """
        field = Field.forTypes("path", [unicode, int], "A path!")
        field.validate(123)
        field.validate("hello")

    def test_validateNone(self):
        """
        When given a "class" of C{None}, L{Field.validate} will support
        validating C{None}.
        """
        field = Field.forTypes("None", [None], "Nothing!")
        field.validate(None)

    def test_validateWrongType(self):
        """
        L{Field.validate} will raise a L{ValidationError} exception if the
        given value's type is not in the list of supported classes.
        """
        field = Field.forTypes("key", [int], "An integer key")
        self.assertRaises(ValidationError, field.validate, "lala")
        self.assertRaises(ValidationError, field.validate, None)
        self.assertRaises(ValidationError, field.validate, object())

    def test_extraValidatorPasses(self):
        """
        L{Field.validate} will not raise an exception if the extra validator
        does not raise an exception.
        """

        def validate(i):
            if i > 10:
                return
            else:
                raise ValidationError("too small")

        field = Field.forTypes("key", [int], "An integer key", validate)
        field.validate(11)

    def test_extraValidatorFails(self):
        """
        L{Field.validate} will raise a L{ValidationError} exception if the
        extra validator raises one.
        """

        def validate(i):
            if i > 10:
                return
            else:
                raise ValidationError("too small")

        field = Field.forTypes("key", [int], "An int", validate)
        self.assertRaises(ValidationError, field.validate, 10)

    def test_onlyValidTypes(self):
        """
        Only JSON supported types can be passed to L{Field.forTypes}.
        """
        self.assertRaises(TypeError, Field.forTypes, "key", [complex], "Oops")

    def test_listIsValidType(self):
        """
        A C{list} is a valid type for L{Field.forTypes}.
        """
        Field.forTypes("key", [list], "Oops")

    def test_dictIsValidType(self):
        """
        A C{dict} is a valid type for L{Field.forTypes}.
        """
        Field.forTypes("key", [dict], "Oops")


class FieldTests(TestCase):
    """
    Tests for L{Field}.
    """

    def test_description(self):
        """
        L{Field.description} stores the passed in description.
        """
        field = Field("path", identity, "A path!")
        self.assertEqual(field.description, "A path!")

    def test_optionalDescription(self):
        """
        L{Field} can be constructed with no description.
        """
        field = Field("path", identity)
        self.assertEqual(field.description, "")

    def test_key(self):
        """
        L{Field.key} stores the passed in field key.
        """
        field = Field("path", identity, "A path!")
        self.assertEqual(field.key, "path")

    def test_serialize(self):
        """
        L{Field.serialize} calls the given serializer function.
        """
        result = []
        Field("key", result.append, "field").serialize(123)
        self.assertEqual(result, [123])

    def test_serializeResult(self):
        """
        L{Field.serialize} returns the result of the given serializer function.
        """
        result = Field("key", lambda obj: 456, "field").serialize(None)
        self.assertEqual(result, 456)

    def test_serializeCallsValidate(self):
        """
        L{Field.validate} calls the serializer, in case that raises an
        exception for the given input.
        """

        class MyException(Exception):
            pass

        def serialize(obj):
            raise MyException()

        field = Field("key", serialize, "")
        self.assertRaises(MyException, field.validate, 123)

    def test_noExtraValidator(self):
        """
        L{Field.validate} doesn't break if there is no extra validator.
        """
        field = Field("key", identity, "")
        field.validate(123)

    def test_extraValidatorPasses(self):
        """
        L{Field.validate} will not raise an exception if the extra validator
        does not raise an exception.
        """

        def validate(i):
            if i > 10:
                return
            else:
                raise ValidationError("too small")

        field = Field("path", identity, "A path!", validate)
        field.validate(11)

    def test_extraValidatorFails(self):
        """
        L{Field.validate} will raise a L{ValidationError} exception if the
        extra validator raises one.
        """

        def validate(i):
            if i > 10:
                return
            else:
                raise ValidationError("too small")

        field = Field("path", identity, "A path!", validate)
        self.assertRaises(ValidationError, field.validate, 10)


class FieldForValueTests(TestCase):
    """
    Tests for L{Field.forValue}.
    """

    def test_forValue(self):
        """
        L{Field.forValue} creates a L{Field} with the given key and description.
        """
        field = Field.forValue("key", None, "description")
        self.assertEqual(field.key, "key")
        self.assertEqual(field.description, "description")

    def test_forValueGoodValue(self):
        """
        The L{Field.forValue}-created L{Field} validates the value it was
        constructed with.
        """
        field = Field.forValue("key", 1234, "description")
        field.validate(1234)

    def test_valueFieldWrongValue(self):
        """
        The L{Field.forValue}-created L{Field} raises a L{ValidationError} for
        different values.
        """
        field = Field.forValue("key", 1234, "description")
        self.assertRaises(ValidationError, field.validate, 5678)

    def test_serialize(self):
        """
        The L{Field.forValue}-created L{Field} returns the given object when
        serializing, regardless of input.

        If the caller is buggy, no need to log garbage if we know what needs
        logging. These bugs will be caught by unit tests, anyway, if author of
        code is doing things correctly.
        """
        field = Field.forValue("key", 1234, "description")
        self.assertEqual(field.serialize(None), 1234)


class FieldsTests(TestCase):
    """
    Tests for L{fields}.
    """

    def test_positional(self):
        """
        L{fields} accepts positional arguments of L{Field} instances and
        combines them with fields specied as keyword arguments.
        """
        a_field = Field("akey", identity)
        l = fields(a_field, another=str)
        self.assertIn(a_field, l)
        self.assertEqual(
            {(type(field), field.key) for field in l},
            {(Field, "akey"), (Field, "another")},
        )

    def test_keys(self):
        """
        L{fields} creates L{Field} instances with the given keys.
        """
        l = fields(key=int, status=str)
        self.assertEqual(
            {(type(field), field.key) for field in l},
            {(Field, "key"), (Field, "status")},
        )

    def test_validTypes(self):
        """
        The L{Field} instances constructed by L{fields} validate the specified
        types.
        """
        (field,) = fields(key=int)
        self.assertRaises(ValidationError, field.validate, "abc")

    def test_noSerialization(self):
        """
        The L{Field} instances constructed by L{fields} do no special
        serialization.
        """
        (field,) = fields(key=int)
        self.assertEqual(field.serialize("abc"), "abc")


class MessageSerializerTests(TestCase):
    """
    Tests for L{_MessageSerializer}.
    """

    def test_noMultipleFields(self):
        """
        L{_MessageSerializer.__init__} will raise a L{ValueError} exception if
        constructed with more than object per field name.
        """
        self.assertRaises(
            ValueError,
            _MessageSerializer,
            [
                Field("akey", identity, ""),
                Field("akey", identity, ""),
                Field("message_type", identity, ""),
            ],
        )

    def test_noBothTypeFields(self):
        """
        L{_MessageSerializer.__init__} will raise a L{ValueError} exception if
        constructed with both a C{"message_type"} and C{"action_type"} field.
        """
        self.assertRaises(
            ValueError,
            _MessageSerializer,
            [Field("message_type", identity, ""), Field("action_type", identity, "")],
        )

    def test_missingTypeField(self):
        """
        L{_MessageSerializer.__init__} will raise a L{ValueError} if there is
        neither a C{"message_type"} nor a C{"action_type"} field.
        """
        self.assertRaises(ValueError, _MessageSerializer, [])

    def test_noTaskLevel(self):
        """
        L{_MessageSerializer.__init__} will raise a L{ValueError} if there is
        a C{"task_level"} field included.
        """
        self.assertRaises(
            ValueError,
            _MessageSerializer,
            [Field("message_type", identity, ""), Field("task_level", identity, "")],
        )

    def test_noTaskUuid(self):
        """
        L{_MessageSerializer.__init__} will raise a L{ValueError} if there is
        a C{"task_uuid"} field included.
        """
        self.assertRaises(
            ValueError,
            _MessageSerializer,
            [Field("message_type", identity, ""), Field("task_uuid", identity, "")],
        )

    def test_noTimestamp(self):
        """
        L{_MessageSerializer.__init__} will raise a L{ValueError} if there is
        a C{"timestamp"} field included.
        """
        self.assertRaises(
            ValueError,
            _MessageSerializer,
            [Field("message_type", identity, ""), Field("timestamp", identity, "")],
        )

    def test_noUnderscoreStart(self):
        """
        L{_MessageSerializer.__init__} will raise a L{ValueError} if there is
        a field included whose name starts with C{"_"}.
        """
        self.assertRaises(
            ValueError,
            _MessageSerializer,
            [Field("message_type", identity, ""), Field("_key", identity, "")],
        )

    def test_serialize(self):
        """
        L{_MessageSerializer.serialize} will serialize all values in the given
        dictionary using the respective L{Field}.
        """
        serializer = _MessageSerializer(
            [
                Field.forValue("message_type", "mymessage", "The type"),
                Field("length", len, "The length of a thing"),
            ]
        )
        message = {"message_type": "mymessage", "length": "thething"}
        serializer.serialize(message)
        self.assertEqual(message, {"message_type": "mymessage", "length": 8})

    def test_missingSerializer(self):
        """
        If a value in the dictionary passed to L{_MessageSerializer.serialize}
        has no respective field, it is unchanged.

        Logging attempts to capture everything, with minimal work; with any
        luck this value is JSON-encodable. Unit tests should catch such bugs, in any case.
        """
        serializer = _MessageSerializer(
            [
                Field.forValue("message_type", "mymessage", "The type"),
                Field("length", len, "The length of a thing"),
            ]
        )
        message = {"message_type": "mymessage", "length": "thething", "extra": 123}
        serializer.serialize(message)
        self.assertEqual(
            message, {"message_type": "mymessage", "length": 8, "extra": 123}
        )

    def test_fieldInstances(self):
        """
        Fields to L{_MessageSerializer.__init__} should be instances of
        L{Field}.
        """
        a_field = Field("a_key", identity)
        arg = object()
        with self.assertRaises(TypeError) as cm:
            _MessageSerializer([a_field, arg])
        self.assertEqual(("Expected a Field instance but got", arg), cm.exception.args)


class MessageTypeTests(TestCase):
    """
    Tests for L{MessageType}.
    """

    def messageType(self):
        """
        Return a L{MessageType} suitable for unit tests.
        """
        return MessageType(
            "myapp:mysystem",
            [Field.forTypes("key", [int], ""), Field.forTypes("value", [int], "")],
            "A message type",
        )

    def test_validateMissingType(self):
        """
        L{MessageType._serializer.validate} raises a L{ValidationError} exception if the
        given dictionary has no C{"message_type"} field.
        """
        messageType = self.messageType()
        self.assertRaises(
            ValidationError, messageType._serializer.validate, {"key": 1, "value": 2}
        )

    def test_validateWrongType(self):
        """
        L{MessageType._serializer.validate} raises a L{ValidationError}
        exception if the given dictionary has the wrong value for the
        C{"message_type"} field.
        """
        messageType = self.messageType()
        self.assertRaises(
            ValidationError,
            messageType._serializer.validate,
            {"key": 1, "value": 2, "message_type": "wrong"},
        )

    def test_validateExtraField(self):
        """
        L{MessageType._serializer.validate} raises a L{ValidationError}
        exception if the given dictionary has an extra unknown field.
        """
        messageType = self.messageType()
        self.assertRaises(
            ValidationError,
            messageType._serializer.validate,
            {"key": 1, "value": 2, "message_type": "myapp:mysystem", "extra": "hello"},
        )

    def test_validateMissingField(self):
        """
        L{MessageType._serializer.validate} raises a L{ValidationError}
        exception if the given dictionary has a missing field.
        """
        messageType = self.messageType()
        self.assertRaises(
            ValidationError,
            messageType._serializer.validate,
            {"key": 1, "message_type": "myapp:mysystem"},
        )

    def test_validateFieldValidation(self):
        """
        L{MessageType._serializer.validate} raises a L{ValidationError}
        exception if the one of the field values fails field-specific
        validation.
        """
        messageType = self.messageType()
        self.assertRaises(
            ValidationError,
            messageType._serializer.validate,
            {"key": 1, "value": None, "message_type": "myapp:mysystem"},
        )

    def test_validateStandardFields(self):
        """
        L{MessageType._serializer.validate} does not raise an exception if the
        dictionary has the standard fields that are added to all messages.
        """
        messageType = self.messageType()
        messageType._serializer.validate(
            {
                "key": 1,
                "value": 2,
                "message_type": "myapp:mysystem",
                "task_level": "/",
                "task_uuid": "123",
                "timestamp": "xxx",
            }
        )

    def test_call(self):
        """
        L{MessageType.__call__} creates a new L{Message} with correct
        C{message_type} field value added.
        """
        messageType = self.messageType()
        message = messageType()
        self.assertEqual(message._contents, {"message_type": messageType.message_type})

    def test_callSerializer(self):
        """
        L{MessageType.__call__} creates a new L{Message} with the
        L{MessageType._serializer} as its serializer.
        """
        messageType = self.messageType()
        message = messageType()
        self.assertIs(message._serializer, messageType._serializer)

    def test_callWithFields(self):
        """
        L{MessageType.__call__} creates a new L{Message} with the additional
        given fields.
        """
        messageType = self.messageType()
        message = messageType(key=2, value=3)
        self.assertEqual(
            message._contents,
            {"message_type": messageType.message_type, "key": 2, "value": 3},
        )

    def test_logCallsDefaultLoggerWrite(self):
        """
        L{MessageType.log} calls the given logger's C{write} method with a
        dictionary that is superset of the L{Message} contents.
        """
        messages = []
        add_destination(messages.append)
        self.addCleanup(remove_destination, messages.append)
        message_type = self.messageType()
        message_type.log(key=1234, value=3)
        self.assertEqual(messages[0]["key"], 1234)
        self.assertEqual(messages[0]["value"], 3)
        self.assertEqual(messages[0]["message_type"], message_type.message_type)

    def test_description(self):
        """
        L{MessageType.description} stores the passed in description.
        """
        messageType = self.messageType()
        self.assertEqual(messageType.description, "A message type")

    def test_optionalDescription(self):
        """
        L{MessageType} can be constructed without a description.
        """
        messageType = MessageType("name", [])
        self.assertEqual(messageType.description, "")


class ActionTypeTestsMixin(object):
    """
    Mixin for tests for the three L{ActionType} message variants.
    """

    def getValidMessage(self):
        """
        Return a dictionary of a message that is of the action status being
        tested.
        """
        raise NotImplementedError("Override in subclasses")

    def getSerializer(self, actionType):
        """
        Given a L{ActionType}, return the L{_MessageSerializer} for this
        variant.
        """
        raise NotImplementedError("Override in subclasses")

    def actionType(self):
        """
        Return a L{ActionType} suitable for unit tests.
        """
        return ActionType(
            "myapp:mysystem:myaction",
            [Field.forTypes("key", [int], "")],  # start fields
            [Field.forTypes("value", [int], "")],  # success fields
            "A action type",
        )

    def test_validateMissingType(self):
        """
        L{ActionType.validate} raises a L{ValidationError} exception if the
        given dictionary has no C{"action_type"} field.
        """
        actionType = self.actionType()
        message = self.getValidMessage()
        del message["action_type"]
        self.assertRaises(
            ValidationError, self.getSerializer(actionType).validate, message
        )

    def test_validateWrongType(self):
        """
        L{ActionType.validate} raises a L{ValidationError} exception if the
        given dictionary has the wrong value for the C{"action_type"} field.
        """
        actionType = self.actionType()
        message = self.getValidMessage()
        message["action_type"] = "xxx"
        self.assertRaises(
            ValidationError, self.getSerializer(actionType).validate, message
        )

    def test_validateExtraField(self):
        """
        L{ActionType.validate} raises a L{ValidationError} exception if the
        given dictionary has an extra unknown field.
        """
        actionType = self.actionType()
        message = self.getValidMessage()
        message["extra"] = "ono"
        self.assertRaises(
            ValidationError, self.getSerializer(actionType).validate, message
        )

    def test_validateMissingField(self):
        """
        L{ActionType.validate} raises a L{ValidationError} exception if the
        given dictionary has a missing field.
        """
        actionType = self.actionType()
        message = self.getValidMessage()
        for key in message:
            if key != "action_type":
                del message[key]
                break
        self.assertRaises(
            ValidationError, self.getSerializer(actionType).validate, message
        )

    def test_validateFieldValidation(self):
        """
        L{ActionType.validate} raises a L{ValidationError} exception if the
        one of the field values fails field-specific validation.
        """
        actionType = self.actionType()
        message = self.getValidMessage()
        for key in message:
            if key != "action_type":
                message[key] = object()
                break
        self.assertRaises(
            ValidationError, self.getSerializer(actionType).validate, message
        )

    def test_validateStandardFields(self):
        """
        L{ActionType.validate} does not raise an exception if the dictionary
        has the standard fields that are added to all messages.
        """
        actionType = self.actionType()
        message = self.getValidMessage()
        message.update({"task_level": "/", "task_uuid": "123", "timestamp": "xxx"})
        self.getSerializer(actionType).validate(message)


class ActionTypeStartMessage(TestCase, ActionTypeTestsMixin):
    """
    Tests for L{ActionType} validation of action start messages.
    """

    def getValidMessage(self):
        """
        Return a dictionary of a valid action start message.
        """
        return {
            "action_type": "myapp:mysystem:myaction",
            "action_status": "started",
            "key": 1,
        }

    def getSerializer(self, actionType):
        return actionType._serializers.start


class ActionTypeSuccessMessage(TestCase, ActionTypeTestsMixin):
    """
    Tests for L{ActionType} validation of action success messages.
    """

    def getValidMessage(self):
        """
        Return a dictionary of a valid action success message.
        """
        return {
            "action_type": "myapp:mysystem:myaction",
            "action_status": "succeeded",
            "value": 2,
        }

    def getSerializer(self, actionType):
        return actionType._serializers.success


class ActionTypeFailureMessage(TestCase, ActionTypeTestsMixin):
    """
    Tests for L{ActionType} validation of action failure messages.
    """

    def getValidMessage(self):
        """
        Return a dictionary of a valid action failure message.
        """
        return {
            "action_type": "myapp:mysystem:myaction",
            "action_status": "failed",
            "exception": "exceptions.RuntimeError",
            "reason": "because",
        }

    def getSerializer(self, actionType):
        return actionType._serializers.failure

    def test_validateExtraField(self):
        """
        Additional fields (which can be added by exception extraction) don't
        cause a validation failure for failed action messages.
        """
        actionType = self.actionType()
        message = self.getValidMessage()
        message.update({"task_level": "/", "task_uuid": "123", "timestamp": "xxx"})
        message.update({"extra_field": "hello"})
        self.getSerializer(actionType).validate(message)


class ChildActionTypeStartMessage(TestCase):
    """
    Tests for validation of child actions created with L{ActionType}.
    """

    def test_childActionUsesChildValidator(self):
        """
        Validation of child actions uses the child's validator.
        """
        A = ActionType("myapp:foo", [Field.forTypes("a", [int], "")], [], "")
        B = ActionType("myapp:bar", [Field.forTypes("b", [int], "")], [], "")

        logger = MemoryLogger()

        with A(logger, a=1):
            with B(logger, b=2):
                pass
        # If wrong serializers/validators were used, this will fail:
        logger.validate()


class ActionTypeTests(TestCase):
    """
    General tests for L{ActionType}.
    """

    def actionType(self):
        """
        Return a L{ActionType} suitable for unit tests.
        """
        return ActionType("myapp:mysystem:myaction", [], [], "An action type")

    def test_call(self):
        """
        L{ActionType.__call__} returns the result of calling
        C{self._start_action}.
        """
        actionType = self.actionType()
        actionType._start_action = lambda *args, **kwargs: 1234
        result = actionType(object())
        self.assertEqual(result, 1234)

    def test_callArguments(self):
        """
        L{ActionType.__call__} calls C{self._start_action} with the logger,
        action type, serializers and passed in fields.
        """
        called = []
        actionType = self.actionType()
        actionType._start_action = lambda *args, **kwargs: called.append((args, kwargs))
        logger = object()
        actionType(logger, key=5)
        self.assertEqual(
            called,
            [
                (
                    (logger, "myapp:mysystem:myaction", actionType._serializers),
                    {"key": 5},
                )
            ],
        )

    def test_defaultStartAction(self):
        """
        L{ActionType._start_action} is L{eliot.start_action} by default.
        """
        self.assertEqual(ActionType._start_action, start_action)

    def test_as_task(self):
        """
        L{ActionType.as_task} returns the result of calling C{self._startTask}.
        """
        actionType = self.actionType()
        actionType._startTask = lambda *args, **kwargs: 1234
        result = actionType.as_task(object())
        self.assertEqual(result, 1234)

    def test_as_taskArguments(self):
        """
        L{ActionType.as_task} calls C{self._startTask} with the logger,
        action type and passed in fields.
        """
        called = []
        actionType = self.actionType()
        actionType._startTask = lambda *args, **kwargs: called.append((args, kwargs))
        logger = object()
        actionType.as_task(logger, key=5)
        self.assertEqual(
            called,
            [
                (
                    (logger, "myapp:mysystem:myaction", actionType._serializers),
                    {"key": 5},
                )
            ],
        )

    def test_defaultStartTask(self):
        """
        L{ActionType._startTask} is L{eliot.startTask} by default.
        """
        self.assertEqual(ActionType._startTask, startTask)

    def test_description(self):
        """
        L{ActionType.description} stores the passed in description.
        """
        actionType = self.actionType()
        self.assertEqual(actionType.description, "An action type")

    def test_optionalDescription(self):
        """
        L{ActionType} can be constructed without a description.
        """
        actionType = ActionType("name", [], [])
        self.assertEqual(actionType.description, "")

    def test_as_taskDefaultLogger(self):
        """
        L{ActionType.as_task} doesn't require passing in a logger.
        """
        actionType = self.actionType()
        actionType.as_task(key=5)


class EndToEndValidationTests(TestCase):
    """
    Test validation of messages created using L{MessageType} and
    L{ActionType}.
    """

    MESSAGE = MessageType(
        "myapp:mymessage",
        [Field.forTypes("key", [int], "The key")],
        "A message for testing.",
    )
    ACTION = ActionType(
        "myapp:myaction",
        [Field.forTypes("key", [int], "The key")],
        [Field.forTypes("result", [unicode], "The result")],
        "An action for testing.",
    )

    def test_correctFromMessageType(self):
        """
        A correct message created using L{MessageType} will be logged to a
        L{MemoryLogger}.
        """
        logger = MemoryLogger()
        msg = self.MESSAGE().bind(key=123)
        msg.write(logger)
        self.assertEqual(logger.messages[0]["key"], 123)

    def test_incorrectFromMessageType(self):
        """
        An incorrect message created using L{MessageType} will raise a
        L{ValidationError} in L{MemoryLogger.validate}.
        """
        logger = MemoryLogger()
        msg = self.MESSAGE().bind(key="123")
        msg.write(logger)
        self.assertRaises(ValidationError, logger.validate)

    def test_correctStartFromActionType(self):
        """
        A correct start message created using a L{ActionType} will be logged
        to a L{MemoryLogger}.
        """
        logger = MemoryLogger()
        with self.ACTION(logger, key=123) as action:
            action.addSuccessFields(result="foo")
        self.assertEqual(logger.messages[0]["key"], 123)

    def test_omitLoggerFromActionType(self):
        """
        If no logger is given to the L{ActionType} the default logger is used.
        """
        messages = []
        add_destination(messages.append)
        self.addCleanup(remove_destination, messages.append)
        with self.ACTION(key=123) as action:
            action.add_success_fields(result="foo")
        self.assertEqual(messages[0]["key"], 123)

    def test_incorrectStartFromActionType(self):
        """
        An incorrect start message created using a L{ActionType} will raise a
        L{ValidationError}.
        """
        logger = MemoryLogger()
        with self.ACTION(logger, key="123") as action:
            action.addSuccessFields(result="foo")
        self.assertRaises(ValidationError, logger.validate)

    def test_correctSuccessFromActionType(self):
        """
        A correct success message created using a L{ActionType} will be logged
        to a L{MemoryLogger}.
        """
        logger = MemoryLogger()
        with self.ACTION(logger, key=123) as action:
            action.addSuccessFields(result="foo")
        self.assertEqual(logger.messages[1]["result"], "foo")

    def test_incorrectSuccessFromActionType(self):
        """
        An incorrect success message created using a L{ActionType} will raise a
        L{ValidationError}.
        """
        logger = MemoryLogger()
        with self.ACTION(logger, key=123) as action:
            action.addSuccessFields(result=-1)
        self.assertRaises(ValidationError, logger.validate)

    def test_correctFailureFromActionType(self):
        """
        A correct failure message created using a L{ActionType} will be logged
        to a L{MemoryLogger}.
        """
        logger = MemoryLogger()

        def run():
            with self.ACTION(logger, key=123):
                raise RuntimeError("hello")

        self.assertRaises(RuntimeError, run)
        self.assertEqual(logger.messages[1]["reason"], "hello")


class PEP8Tests(TestCase):
    """
    Tests for PEP 8 method compatibility.
    """

    def test_for_value(self):
        """
        L{Field.for_value} is the same as L{Field.forValue}.
        """
        self.assertEqual(Field.for_value, Field.forValue)

    def test_for_types(self):
        """
        L{Field.for_types} is the same as L{Field.forTypes}.
        """
        self.assertEqual(Field.for_types, Field.forTypes)

    def test_as_task(self):
        """
        L{ActionType.as_task} is the same as L{ActionType.asTask}.
        """
        self.assertEqual(ActionType.as_task, ActionType.asTask)



"""
Tests for L{eliot._message}.
"""

from __future__ import unicode_literals

from unittest import TestCase
from uuid import UUID
import time

from pyrsistent import pmap

try:
    from twisted.python.failure import Failure
except ImportError:
    Failure = None

from .._message import WrittenMessage, Message, log_message
from .._output import MemoryLogger
from .._action import Action, start_action, TaskLevel
from .. import add_destinations, remove_destination


class DeprecatedMessageTests(TestCase):
    """
    Test for L{Message}.
    """

    def test_new(self):
        """
        L{Message.new} returns a new L{Message} that is initialized with the
        given keyword arguments as its contents, and a default message type.
        """
        msg = Message.new(key="value", another=2)
        self.assertEqual(msg.contents(), {"key": "value", "another": 2})

    def test_contentsCopies(self):
        """
        L{Message.contents} returns a copy of the L{Message} contents.
        """
        msg = Message.new(key="value")
        del msg.contents()["key"]
        self.assertEqual(msg.contents(), {"key": "value"})

    def test_bindOverwrites(self):
        """
        L{Message.bind} returns a new L{Message} whose contents include the
        additional given fields.
        """
        msg = Message.new(key="value", another=2)
        another = msg.bind(another=3, more=4)
        self.assertIsInstance(another, Message)
        self.assertEqual(another.contents(), {"key": "value", "another": 3, "more": 4})

    def test_bindPreservesOriginal(self):
        """
        L{Message.bind} does not mutate the instance it is called on.
        """
        msg = Message.new(key=4)
        msg.bind(key=6)
        self.assertEqual(msg.contents(), {"key": 4})

    def test_writeCallsLoggerWrite(self):
        """
        L{Message.write} calls the given logger's C{write} method with a
        dictionary that is superset of the L{Message} contents.
        """
        logger = MemoryLogger()
        msg = Message.new(key=4)
        msg.write(logger)
        self.assertEqual(len(logger.messages), 1)
        self.assertEqual(logger.messages[0]["key"], 4)

    def test_writeDefaultLogger(self):
        """
        L{Message.write} writes to the default logger if none is given.
        """
        messages = []
        add_destinations(messages.append)
        self.addCleanup(remove_destination, messages.append)
        Message.new(some_key=1234).write()
        self.assertEqual(messages[0]["some_key"], 1234)

    def test_writeCreatesNewDictionary(self):
        """
        L{Message.write} creates a new dictionary on each call.

        This is important because we mutate the dictionary in
        ``Logger.write``, so we want to make sure the ``Message`` is unchanged
        in that case. In general we want ``Message`` objects to be effectively
        immutable.
        """

        class Logger(list):
            def write(self, d, serializer):
                self.append(d)

        logger = Logger()
        msg = Message.new(key=4)
        msg.write(logger)
        logger[0]["key"] = 5
        msg.write(logger)
        self.assertEqual(logger[1]["key"], 4)

    def test_logCallsDefaultLoggerWrite(self):
        """
        L{Message.log} calls the default logger's C{write} method with a
        dictionary that is superset of the L{Message} contents.
        """
        messages = []
        add_destinations(messages.append)
        self.addCleanup(remove_destination, messages.append)
        Message.log(some_key=1234)
        self.assertEqual(messages[0]["some_key"], 1234)

    def test_defaultTime(self):
        """
        L{Message._time} returns L{time.time} by default.
        """
        msg = Message({})
        self.assertIs(msg._time, time.time)

    def test_writeAddsTimestamp(self):
        """
        L{Message.write} adds a C{"timestamp"} field to the dictionary written
        to the logger, with the current time in seconds since the epoch.
        """
        logger = MemoryLogger()
        msg = Message.new(key=4)
        msg.write(logger)
        self.assertTrue(time.time() - logger.messages[0]["timestamp"] < 0.1)

    def test_write_preserves_message_type(self):
        """
        L{Message.write} doesn't add a C{message_type} if one is already set.
        """
        logger = MemoryLogger()
        msg = Message.new(key=4, message_type="isetit")
        msg.write(logger)
        self.assertEqual(logger.messages[0]["message_type"], "isetit")
        self.assertNotIn("action_type", logger.messages[0])

    def test_explicitAction(self):
        """
        L{Message.write} adds the identification fields from the given
        L{Action} to the dictionary written to the logger, as well as a
        message_type if none is set.
        """
        logger = MemoryLogger()
        action = Action(logger, "unique", TaskLevel(level=[]), "sys:thename")
        msg = Message.new(key=2)
        msg.write(logger, action)
        written = logger.messages[0]
        del written["timestamp"]
        self.assertEqual(
            written,
            {"task_uuid": "unique", "task_level": [1], "key": 2, "message_type": ""},
        )

    def test_implicitAction(self):
        """
        If no L{Action} is specified, L{Message.write} adds the identification
        fields from the current execution context's L{Action} to the
        dictionary written to the logger.
        """
        logger = MemoryLogger()
        action = Action(logger, "unique", TaskLevel(level=[]), "sys:thename")
        msg = Message.new(key=2)
        with action:
            msg.write(logger)
        written = logger.messages[0]
        del written["timestamp"]
        self.assertEqual(
            written,
            {"task_uuid": "unique", "task_level": [1], "key": 2, "message_type": ""},
        )

    def test_missingAction(self):
        """
        If no L{Action} is specified, and the current execution context has no
        L{Action}, a new task_uuid is generated.

        This ensures all messages have a unique identity, as specified by
        task_uuid/task_level.
        """
        logger = MemoryLogger()
        Message.new(key=2).write(logger)
        Message.new(key=3).write(logger)

        message1, message2 = logger.messages
        self.assertEqual(
            (
                UUID(message1["task_uuid"]) != UUID(message2["task_uuid"]),
                message1["task_level"],
                message2["task_level"],
            ),
            (True, [1], [1]),
        )

    def test_actionCounter(self):
        """
        Each message written within the context of an L{Action} gets its
        C{task_level} field incremented.
        """
        logger = MemoryLogger()
        msg = Message.new(key=2)
        with start_action(logger, "sys:thename"):
            for i in range(4):
                msg.write(logger)
        # We expect 6 messages: start action, 4 standalone messages, finish
        # action:
        self.assertEqual(
            [m["task_level"] for m in logger.messages], [[1], [2], [3], [4], [5], [6]]
        )

    def test_writePassesSerializer(self):
        """
        If a L{Message} is created with a serializer, it is passed as a second
        argument to the logger when C{write} is called.
        """

        class ListLogger(list):
            def write(self, dictionary, serializer):
                self.append(serializer)

        logger = ListLogger()
        serializer = object()
        msg = Message({}, serializer)
        msg.write(logger)
        self.assertIs(logger[0], serializer)

    def test_serializerPassedInBind(self):
        """
        The L{Message} returned by L{Message.bind} includes the serializer
        passed to the parent.
        """
        serializer = object()
        msg = Message({}, serializer)
        msg2 = msg.bind(x=1)
        self.assertIs(msg2._serializer, serializer)

    def test_newWithSerializer(self):
        """
        L{Message.new} can accept a serializer.
        """
        serializer = object()
        msg = Message.new(serializer, x=1)
        self.assertIs(msg._serializer, serializer)


class WrittenMessageTests(TestCase):
    """
    Tests for L{WrittenMessage}.
    """

    def test_as_dict(self):
        """
        L{WrittenMessage.as_dict} returns the dictionary that will be serialized
        to the log.
        """
        log_entry = pmap(
            {"timestamp": 1, "task_uuid": "unique", "task_level": [1], "foo": "bar"}
        )
        self.assertEqual(WrittenMessage.from_dict(log_entry).as_dict(), log_entry)

    def test_from_dict(self):
        """
        L{WrittenMessage.from_dict} converts a dictionary that has been
        deserialized from a log into a L{WrittenMessage} object.
        """
        log_entry = pmap(
            {"timestamp": 1, "task_uuid": "unique", "task_level": [1], "foo": "bar"}
        )
        parsed = WrittenMessage.from_dict(log_entry)
        self.assertEqual(parsed.timestamp, 1)
        self.assertEqual(parsed.task_uuid, "unique")
        self.assertEqual(parsed.task_level, TaskLevel(level=[1]))
        self.assertEqual(parsed.contents, pmap({"foo": "bar"}))


class LogMessageTests(TestCase):
    """Tests for L{log_message}."""

    def test_writes_message(self):
        """
        L{log_message} writes to the default logger.
        """
        messages = []
        add_destinations(messages.append)
        self.addCleanup(remove_destination, messages.append)
        log_message(message_type="hello", some_key=1234)
        self.assertEqual(messages[0]["some_key"], 1234)
        self.assertEqual(messages[0]["message_type"], "hello")
        self.assertTrue(time.time() - messages[0]["timestamp"] < 0.1)

    def test_implicitAction(self):
        """
        If no L{Action} is specified, L{log_message} adds the identification
        fields from the current execution context's L{Action} to the
        dictionary written to the logger.
        """
        logger = MemoryLogger()
        action = Action(logger, "unique", TaskLevel(level=[]), "sys:thename")
        with action:
            log_message(key=2, message_type="a")
        written = logger.messages[0]
        del written["timestamp"]
        self.assertEqual(
            written,
            {"task_uuid": "unique", "task_level": [1], "key": 2, "message_type": "a"},
        )

    def test_missingAction(self):
        """
        If no L{Action} is specified, and the current execution context has no
        L{Action}, a new task_uuid is generated.

        This ensures all messages have a unique identity, as specified by
        task_uuid/task_level.
        """
        messages = []
        add_destinations(messages.append)
        self.addCleanup(remove_destination, messages.append)

        log_message(key=2, message_type="")
        log_message(key=3, message_type="")

        message1, message2 = messages
        self.assertEqual(
            (
                UUID(message1["task_uuid"]) != UUID(message2["task_uuid"]),
                message1["task_level"],
                message2["task_level"],
            ),
            (True, [1], [1]),
        )



"""
Tests for L{eliot.journald}.
"""
from os import getpid, strerror
from unittest import skipUnless, TestCase
from subprocess import check_output, CalledProcessError, STDOUT
from errno import EINVAL
from sys import argv
from uuid import uuid4
from time import sleep

from six import text_type as unicode

from .._bytesjson import loads
from .._output import MemoryLogger
from .._message import TASK_UUID_FIELD
from .. import start_action, Message, write_traceback

try:
    from ..journald import sd_journal_send, JournaldDestination
except ImportError:
    sd_journal_send = None


def _journald_available():
    """
    :return: Boolean indicating whether journald is available to use.
    """
    if sd_journal_send is None:
        return False
    try:
        check_output(["journalctl", "-b", "-n1"], stderr=STDOUT)
    except (OSError, CalledProcessError):
        return False
    return True


def last_journald_message():
    """
    @return: Last journald message from this process as a dictionary in
         journald JSON format.
    """
    # It may take a little for messages to actually reach journald, so we
    # write out marker message and wait until it arrives. We can then be
    # sure the message right before it is the one we want.
    marker = unicode(uuid4())
    sd_journal_send(MESSAGE=marker.encode("ascii"))
    for i in range(500):
        messages = check_output(
            [
                b"journalctl",
                b"-a",
                b"-o",
                b"json",
                b"-n2",
                b"_PID=" + str(getpid()).encode("ascii"),
            ]
        )
        messages = [loads(m) for m in messages.splitlines()]
        if len(messages) == 2 and messages[1]["MESSAGE"] == marker:
            return messages[0]
        sleep(0.01)
    raise RuntimeError("Message never arrived?!")


class SdJournaldSendTests(TestCase):
    """
    Functional tests for L{sd_journal_send}.
    """

    @skipUnless(
        _journald_available(), "journald unavailable or inactive on this machine."
    )
    def setUp(self):
        pass

    def assert_roundtrip(self, value):
        """
        Write a value as a C{MESSAGE} field, assert it is output.

        @param value: Value to write as unicode.
        """
        sd_journal_send(MESSAGE=value)
        result = last_journald_message()
        self.assertEqual(value, result["MESSAGE"].encode("utf-8"))

    def test_message(self):
        """
        L{sd_journal_send} can write a C{MESSAGE} field.
        """
        self.assert_roundtrip(b"hello")

    def test_percent(self):
        """
        L{sd_journal_send} can write a C{MESSAGE} field with a percent.

        Underlying C API calls does printf formatting so this is a
        plausible failure mode.
        """
        self.assert_roundtrip(b"hello%world")

    def test_large(self):
        """
        L{sd_journal_send} can write a C{MESSAGE} field with a large message.
        """
        self.assert_roundtrip(b"hello world" * 20000)

    def test_multiple_fields(self):
        """
        L{sd_journal_send} can send multiple fields.
        """
        sd_journal_send(MESSAGE=b"hello", BONUS_FIELD=b"world")
        result = last_journald_message()
        self.assertEqual(
            (b"hello", b"world"),
            (result["MESSAGE"].encode("ascii"), result["BONUS_FIELD"].encode("ascii")),
        )

    def test_error(self):
        """
        L{sd_journal_send} raises an error when it gets a non-0 result
        from the underlying API.
        """
        with self.assertRaises(IOError) as context:
            sd_journal_send(**{"": b"123"})
        exc = context.exception
        self.assertEqual((exc.errno, exc.strerror), (EINVAL, strerror(EINVAL)))


class JournaldDestinationTests(TestCase):
    """
    Tests for L{JournaldDestination}.
    """

    @skipUnless(
        _journald_available(), "journald unavailable or inactive on this machine."
    )
    def setUp(self):
        self.destination = JournaldDestination()
        self.logger = MemoryLogger()

    def test_json(self):
        """
        The message is stored as JSON in the MESSAGE field.
        """
        Message.new(hello="world", key=123).write(self.logger)
        message = self.logger.messages[0]
        self.destination(message)
        self.assertEqual(loads(last_journald_message()["MESSAGE"]), message)

    def assert_field_for(self, message, field_name, field_value):
        """
        If the given message is logged by Eliot, the given journald field has
        the expected value.

        @param message: Dictionary to log.
        @param field_name: Journald field name to check.
        @param field_value: Expected value for the field.
        """
        self.destination(message)
        self.assertEqual(last_journald_message()[field_name], field_value)

    def test_action_type(self):
        """
        The C{action_type} is stored in the ELIOT_TYPE field.
        """
        action_type = "test:type"
        start_action(self.logger, action_type=action_type)
        self.assert_field_for(self.logger.messages[0], "ELIOT_TYPE", action_type)

    def test_message_type(self):
        """
        The C{message_type} is stored in the ELIOT_TYPE field.
        """
        message_type = "test:type:message"
        Message.new(message_type=message_type).write(self.logger)
        self.assert_field_for(self.logger.messages[0], "ELIOT_TYPE", message_type)

    def test_no_type(self):
        """
        An empty string is stored in ELIOT_TYPE if no type is known.
        """
        Message.new().write(self.logger)
        self.assert_field_for(self.logger.messages[0], "ELIOT_TYPE", "")

    def test_uuid(self):
        """
        The task UUID is stored in the ELIOT_TASK field.
        """
        start_action(self.logger, action_type="xxx")
        self.assert_field_for(
            self.logger.messages[0],
            "ELIOT_TASK",
            self.logger.messages[0][TASK_UUID_FIELD],
        )

    def test_info_priorities(self):
        """
        Untyped messages, action start, successful action end, random typed
        message all get priority 6 ("info").
        """
        with start_action(self.logger, action_type="xxx"):
            Message.new(message_type="msg").write(self.logger)
            Message.new(x=123).write(self.logger)
        priorities = []
        for message in self.logger.messages:
            self.destination(message)
            priorities.append(last_journald_message()["PRIORITY"])
        self.assertEqual(priorities, ["6", "6", "6", "6"])

    def test_error_priority(self):
        """
        A failed action gets priority 3 ("error").
        """
        try:
            with start_action(self.logger, action_type="xxx"):
                raise ZeroDivisionError()
        except ZeroDivisionError:
            pass
        self.assert_field_for(self.logger.messages[-1], "PRIORITY", "3")

    def test_critical_priority(self):
        """
        A traceback gets priority 2 ("critical").
        """
        try:
            raise ZeroDivisionError()
        except ZeroDivisionError:
            write_traceback(logger=self.logger)
        self.assert_field_for(self.logger.serialize()[-1], "PRIORITY", "2")

    def test_identifier(self):
        """
        C{SYSLOG_IDENTIFIER} defaults to C{os.path.basename(sys.argv[0])}.
        """
        identifier = "/usr/bin/testing123"
        try:
            original = argv[0]
            argv[0] = identifier
            # Recreate JournaldDestination with the newly set argv[0].
            self.destination = JournaldDestination()
            Message.new(message_type="msg").write(self.logger)
            self.assert_field_for(
                self.logger.messages[0], "SYSLOG_IDENTIFIER", "testing123"
            )
        finally:
            argv[0] = original



"""
APIs for using Eliot from Twisted.
"""

from __future__ import absolute_import, unicode_literals

import os
import sys

from twisted.logger import Logger as TwistedLogger
from twisted.python.failure import Failure
from twisted.internet.defer import inlineCallbacks

from ._action import current_action
from . import addDestination
from ._generators import eliot_friendly_generator_function

__all__ = [
    "AlreadyFinished",
    "DeferredContext",
    "redirectLogsForTrial",
    "inline_callbacks",
]


def _passthrough(result):
    return result


class AlreadyFinished(Exception):
    """
    L{DeferredContext.addCallbacks} or similar method was called after
    L{DeferredContext.addActionFinish}.

    This indicates a programming bug, e.g. forgetting to unwrap the
    underlying L{Deferred} when passing on to some other piece of code that
    doesn't care about the action context.
    """


class DeferredContext(object):
    """
    A L{Deferred} equivalent of L{eliot.Action.context} and
    L{eliot.action.finish}.

    Makes a L{Deferred}'s callbacks run in a L{eliot.Action}'s context, and
    allows indicating which callbacks to wait for before the action is
    finished.

    The action to use will be taken from the call context.

    @ivar result: The wrapped L{Deferred}.
    """

    def __init__(self, deferred):
        """
        @param deferred: L{twisted.internet.defer.Deferred} to wrap.
        """
        self.result = deferred
        self._action = current_action()
        self._finishAdded = False
        if self._action is None:
            raise RuntimeError(
                "DeferredContext() should only be created in the context of "
                "an eliot.Action."
            )

    def addCallbacks(
        self,
        callback,
        errback=None,
        callbackArgs=None,
        callbackKeywords=None,
        errbackArgs=None,
        errbackKeywords=None,
    ):
        """
        Add a pair of callbacks that will be run in the context of an eliot
        action.

        @return: C{self}
        @rtype: L{DeferredContext}

        @raises AlreadyFinished: L{DeferredContext.addActionFinish} has been
            called. This indicates a programmer error.
        """
        if self._finishAdded:
            raise AlreadyFinished()

        if errback is None:
            errback = _passthrough

        def callbackWithContext(*args, **kwargs):
            return self._action.run(callback, *args, **kwargs)

        def errbackWithContext(*args, **kwargs):
            return self._action.run(errback, *args, **kwargs)

        self.result.addCallbacks(
            callbackWithContext,
            errbackWithContext,
            callbackArgs,
            callbackKeywords,
            errbackArgs,
            errbackKeywords,
        )
        return self

    def addCallback(self, callback, *args, **kw):
        """
        Add a success callback that will be run in the context of an eliot
        action.

        @return: C{self}
        @rtype: L{DeferredContext}

        @raises AlreadyFinished: L{DeferredContext.addActionFinish} has been
            called. This indicates a programmer error.
        """
        return self.addCallbacks(
            callback, _passthrough, callbackArgs=args, callbackKeywords=kw
        )

    def addErrback(self, errback, *args, **kw):
        """
        Add a failure callback that will be run in the context of an eliot
        action.

        @return: C{self}
        @rtype: L{DeferredContext}

        @raises AlreadyFinished: L{DeferredContext.addActionFinish} has been
            called. This indicates a programmer error.
        """
        return self.addCallbacks(
            _passthrough, errback, errbackArgs=args, errbackKeywords=kw
        )

    def addBoth(self, callback, *args, **kw):
        """
        Add a single callback as both success and failure callbacks.

        @return: C{self}
        @rtype: L{DeferredContext}

        @raises AlreadyFinished: L{DeferredContext.addActionFinish} has been
            called. This indicates a programmer error.
        """
        return self.addCallbacks(callback, callback, args, kw, args, kw)

    def addActionFinish(self):
        """
        Indicates all callbacks that should run within the action's context
        have been added, and that the action should therefore finish once
        those callbacks have fired.

        @return: The wrapped L{Deferred}.

        @raises AlreadyFinished: L{DeferredContext.addActionFinish} has been
            called previously. This indicates a programmer error.
        """
        if self._finishAdded:
            raise AlreadyFinished()
        self._finishAdded = True

        def done(result):
            if isinstance(result, Failure):
                exception = result.value
            else:
                exception = None
            self._action.finish(exception)
            return result

        self.result.addBoth(done)
        return self.result


class TwistedDestination(object):
    """
    An Eliot logging destination that forwards logs to Twisted's logging.

    Do not use if you're also redirecting Twisted's logs to Eliot, since then
    you'll have an infinite loop.
    """

    def __init__(self):
        self._logger = TwistedLogger(namespace="eliot")

    def __call__(self, message):
        """
        Log an Eliot message to Twisted's log.

        @param message: A rendered Eliot message.
        @type message: L{dict}
        """
        if message.get("message_type") == "eliot:traceback":
            method = self._logger.critical
        else:
            method = self._logger.info
        method(format="Eliot message: {eliot}", eliot=message)


class _RedirectLogsForTrial(object):
    """
    When called inside a I{trial} process redirect Eliot log messages to
    Twisted's logging system, otherwise do nothing.

    This allows reading Eliot logs output by running unit tests with
    I{trial} in its normal log location: C{_trial_temp/test.log}.

    The way you use it is by calling it a module level in some module that will
    be loaded by trial, typically the top-level C{__init__.py} of your package.

    This function can usually be safely called in all programs since it will
    have no side-effects if used outside of trial. The only exception is you
    are redirecting Twisted logs to Eliot; you should make sure not call
    this function in that case so as to prevent infinite loops. In addition,
    calling the function multiple times has the same effect as calling it
    once.

    (This is not thread-safe at the moment, so in theory multiple threads
    calling this might result in multiple destinatios being added - see
    https://github.com/itamarst/eliot/issues/78).

    Currently this works by checking if C{sys.argv[0]} is called C{trial};
    the ideal mechanism would require
    https://twistedmatrix.com/trac/ticket/6939 to be fixed, but probably
    there are better solutions even without that -
    https://github.com/itamarst/eliot/issues/76 covers those.

    @ivar _sys: An object similar to, and typically identical to, Python's
        L{sys} module.

    @ivar _redirected: L{True} if trial logs have been redirected once already.
    """

    def __init__(self, sys):
        self._sys = sys
        self._redirected = False

    def __call__(self):
        """
        Do the redirect if necessary.

        @return: The destination added to Eliot if any, otherwise L{None}.
        """
        if os.path.basename(self._sys.argv[0]) == "trial" and not self._redirected:
            self._redirected = True
            destination = TwistedDestination()
            addDestination(destination)
            return destination


redirectLogsForTrial = _RedirectLogsForTrial(sys)


def inline_callbacks(original, debug=False):
    """
    Decorate a function like ``inlineCallbacks`` would but in a more
    Eliot-friendly way.  Use it just like ``inlineCallbacks`` but where you
    want Eliot action contexts to Do The Right Thing inside the decorated
    function.
    """
    f = eliot_friendly_generator_function(original)
    if debug:
        f.debug = True
    return inlineCallbacks(f)



"""
journald support for Eliot.
"""

from cffi import FFI
from os import strerror
from sys import argv
from os.path import basename

from ._bytesjson import dumps
from ._message import TASK_UUID_FIELD, MESSAGE_TYPE_FIELD
from ._action import ACTION_TYPE_FIELD, ACTION_STATUS_FIELD, FAILED_STATUS

_ffi = FFI()
_ffi.cdef(
    """
int sd_journal_send(const char *format, ...);
"""
)
try:
    try:
        _journald = _ffi.dlopen("libsystemd.so.0")
    except OSError:
        # Older versions of systemd have separate library:
        _journald = _ffi.dlopen("libsystemd-journal.so.0")
except OSError as e:
    raise ImportError("Failed to load journald: " + str(e))


def sd_journal_send(**kwargs):
    """
    Send a message to the journald log.

    @param kwargs: Mapping between field names to values, both as bytes.

    @raise IOError: If the operation failed.
    """
    # The function uses printf formatting, so we need to quote
    # percentages.
    fields = [
        _ffi.new("char[]", key.encode("ascii") + b"=" + value.replace(b"%", b"%%"))
        for key, value in kwargs.items()
    ]
    fields.append(_ffi.NULL)
    result = _journald.sd_journal_send(*fields)
    if result != 0:
        raise IOError(-result, strerror(-result))


class JournaldDestination(object):
    """
    A logging destination that writes to journald.

    The message will be logged as JSON, with an additional field
    C{ELIOT_TASK} storing the C{task_uuid} and C{ELIOT_TYPE} storing the
    C{message_type} or C{action_type}.

    Messages for failed actions will get priority 3 ("error"), and
    traceback messages will get priority 2 ("critical"). All other
    messages will get priority 1 ("info").
    """

    def __init__(self):
        self._identifier = basename(argv[0]).encode("utf-8")

    def __call__(self, message):
        """
        Write the given message to journald.

        @param message: Dictionary passed from a C{Logger}.
        """
        eliot_type = ""
        priority = b"6"
        if ACTION_TYPE_FIELD in message:
            eliot_type = message[ACTION_TYPE_FIELD]
            if message[ACTION_STATUS_FIELD] == FAILED_STATUS:
                priority = b"3"
        elif MESSAGE_TYPE_FIELD in message:
            eliot_type = message[MESSAGE_TYPE_FIELD]
            if eliot_type == "eliot:traceback":
                priority = b"2"
        sd_journal_send(
            MESSAGE=dumps(message),
            ELIOT_TASK=message[TASK_UUID_FIELD].encode("utf-8"),
            ELIOT_TYPE=eliot_type.encode("utf-8"),
            SYSLOG_IDENTIFIER=self._identifier,
            PRIORITY=priority,
        )



"""
Parse a stream of serialized messages into a forest of
``WrittenAction`` and ``WrittenMessage`` objects.
"""

from __future__ import unicode_literals

from six import text_type as unicode

from pyrsistent import PClass, pmap_field, pset_field, discard

from ._message import WrittenMessage, TASK_UUID_FIELD
from ._action import (
    TaskLevel,
    WrittenAction,
    ACTION_STATUS_FIELD,
    STARTED_STATUS,
    ACTION_TYPE_FIELD,
)


class Task(PClass):
    """
    A tree of actions with the same task UUID.
    """

    _nodes = pmap_field(TaskLevel, (WrittenAction, WrittenMessage))
    _completed = pset_field(TaskLevel)
    _root_level = TaskLevel(level=[])

    def root(self):
        """
        @return: The root L{WrittenAction}.
        """
        return self._nodes[self._root_level]

    def is_complete(self):
        """
        @return bool: True only if all messages in the task tree have been
        added to it.
        """
        return self._root_level in self._completed

    def _insert_action(self, node):
        """
        Add a L{WrittenAction} to the tree.

        Parent actions will be created as necessary.

        @param child: A L{WrittenAction} to add to the tree.

        @return: Updated L{Task}.
        """
        task = self
        if (
            node.end_message
            and node.start_message
            and (len(node.children) == node.end_message.task_level.level[-1] - 2)
        ):
            # Possibly this action is complete, make sure all sub-actions
            # are complete:
            completed = True
            for child in node.children:
                if (
                    isinstance(child, WrittenAction)
                    and child.task_level not in self._completed
                ):
                    completed = False
                    break
            if completed:
                task = task.transform(["_completed"], lambda s: s.add(node.task_level))
        task = task.transform(["_nodes", node.task_level], node)
        return task._ensure_node_parents(node)

    def _ensure_node_parents(self, child):
        """
        Ensure the node (WrittenAction/WrittenMessage) is referenced by parent
        nodes.

        Parent actions will be created as necessary.

        @param child: A L{WrittenMessage} or L{WrittenAction} which is
            being added to the tree.

        @return: Updated L{Task}.
        """
        task_level = child.task_level
        if task_level.parent() is None:
            return self

        parent = self._nodes.get(task_level.parent())
        if parent is None:
            parent = WrittenAction(
                task_level=task_level.parent(), task_uuid=child.task_uuid
            )
        parent = parent._add_child(child)
        return self._insert_action(parent)

    def add(self, message_dict):
        """
        Update the L{Task} with a dictionary containing a serialized Eliot
        message.

        @param message_dict: Dictionary whose task UUID matches this one.

        @return: Updated L{Task}.
        """
        is_action = message_dict.get(ACTION_TYPE_FIELD) is not None
        written_message = WrittenMessage.from_dict(message_dict)
        if is_action:
            action_level = written_message.task_level.parent()
            action = self._nodes.get(action_level)
            if action is None:
                action = WrittenAction(
                    task_level=action_level, task_uuid=message_dict[TASK_UUID_FIELD]
                )
            if message_dict[ACTION_STATUS_FIELD] == STARTED_STATUS:
                # Either newly created MissingAction, or one created by
                # previously added descendant of the action.
                action = action._start(written_message)
            else:
                action = action._end(written_message)
            return self._insert_action(action)
        else:
            # Special case where there is no action:
            if written_message.task_level.level == [1]:
                return self.transform(
                    ["_nodes", self._root_level],
                    written_message,
                    ["_completed"],
                    lambda s: s.add(self._root_level),
                )
            else:
                return self._ensure_node_parents(written_message)


class Parser(PClass):
    """
    Parse serialized Eliot messages into L{Task} instances.

    @ivar _tasks: Map from UUID to corresponding L{Task}.
    """

    _tasks = pmap_field(unicode, Task)

    def add(self, message_dict):
        """
        Update the L{Parser} with a dictionary containing a serialized Eliot
        message.

        @param message_dict: Dictionary of serialized Eliot message.

        @return: Tuple of (list of completed L{Task} instances, updated
            L{Parser}).
        """
        uuid = message_dict[TASK_UUID_FIELD]
        if uuid in self._tasks:
            task = self._tasks[uuid]
        else:
            task = Task()
        task = task.add(message_dict)
        if task.is_complete():
            parser = self.transform(["_tasks", uuid], discard)
            return [task], parser
        else:
            parser = self.transform(["_tasks", uuid], task)
            return [], parser

    def incomplete_tasks(self):
        """
        @return: List of L{Task} that are not yet complete.
        """
        return list(self._tasks.values())

    @classmethod
    def parse_stream(cls, iterable):
        """
        Parse a stream of messages into a stream of L{Task} instances.

        :param iterable: An iterable of serialized Eliot message dictionaries.

        :return: An iterable of parsed L{Task} instances. Remaining
            incomplete L{Task} will be returned when the input stream is
            exhausted.
        """
        parser = Parser()
        for message_dict in iterable:
            completed, parser = parser.add(message_dict)
            for task in completed:
                yield task
        for task in parser.incomplete_tasks():
            yield task


__all__ = ["Parser", "Task", "TaskLevel", "WrittenMessage", "WrittenAction"]



"""
Tests for the eliot package.
"""

# Increase hypothesis deadline so we don't time out on PyPy:
from hypothesis import settings

settings.register_profile("eliot", deadline=1000)
settings.load_profile("eliot")



"""
Python 2/3 JSON encoding/decoding, emulating Python 2's json module.

Python 3 json module doesn't support decoding bytes or encoding. Rather than
adding isinstance checks in main code path which would slow down Python 2,
instead we write our encoder that can support those.
"""

from __future__ import absolute_import

import json as pyjson
import warnings

from six import PY2


def _loads(s):
    """
    Support decoding bytes.
    """
    if isinstance(s, bytes):
        s = s.decode("utf-8")
    return pyjson.loads(s)


def _dumps(obj, cls=pyjson.JSONEncoder):
    """
    Encode to bytes, and presume bytes in inputs are UTF-8 encoded strings.
    """

    class WithBytes(cls):
        """
        JSON encoder that supports L{bytes}.
        """

        def default(self, o):
            if isinstance(o, bytes):
                warnings.warn(
                    "Eliot will soon stop supporting encoding bytes in JSON"
                    " on Python 3",
                    DeprecationWarning,
                )
                return o.decode("utf-8")
            return cls.default(self, o)

    return pyjson.dumps(obj, cls=WithBytes).encode("utf-8")


if PY2:
    # No need for the above on Python 2
    loads, dumps = pyjson.loads, pyjson.dumps
else:
    loads, dumps = _loads, _dumps

__all__ = ["loads", "dumps"]



"""
Common testing infrastructure.
"""

from io import BytesIO
from json import JSONEncoder


class CustomObject(object):
    """Gets encoded to JSON."""


class CustomJSONEncoder(JSONEncoder):
    """JSONEncoder that knows about L{CustomObject}."""

    def default(self, o):
        if isinstance(o, CustomObject):
            return "CUSTOM!"
        return JSONEncoder.default(self, o)


class FakeSys(object):
    """
    A fake L{sys} module.
    """

    def __init__(self, argv, stdinBytes):
        """
        @param argv: List of command-line arguments.

        @param stdinBytes: C{bytes} that are readable from stdin.
        """
        self.argv = argv
        self.stdin = BytesIO(stdinBytes)
        self.stdout = BytesIO()
        self.stderr = BytesIO()



"""
A log message serialization and validation system for Eliot.

Validation is intended to be done by unit tests, not the production code path,
although in theory it could be done then as well.
"""

from __future__ import unicode_literals

from warnings import warn

import six

unicode = six.text_type

from pyrsistent import PClass, field as pyrsistent_field

from ._message import (
    Message,
    REASON_FIELD,
    MESSAGE_TYPE_FIELD,
    TASK_LEVEL_FIELD,
    TASK_UUID_FIELD,
    TIMESTAMP_FIELD,
)
from ._action import (
    start_action,
    startTask,
    ACTION_STATUS_FIELD,
    ACTION_TYPE_FIELD,
    STARTED_STATUS,
    SUCCEEDED_STATUS,
    FAILED_STATUS,
    log_message,
)


class ValidationError(Exception):
    """
    A field value failed validation.
    """


# Types that can be encoded to JSON:
_JSON_TYPES = {type(None), int, float, unicode, list, dict, bytes, bool}
_JSON_TYPES |= set(six.integer_types)

RESERVED_FIELDS = (TASK_LEVEL_FIELD, TASK_UUID_FIELD, TIMESTAMP_FIELD)


class Field(object):
    """
    A named field that can accept rich types and serialize them to the logging
    system's basic types (currently, JSON types).

    An optional extra validation function can be used to validate inputs when
    unit testing.

    @ivar key: The name of the field, the key which refers to it,
        e.g. C{"path"}.

    @ivar description: A description of what this field contains.
    @type description: C{unicode}
    """

    def __init__(self, key, serializer, description="", extraValidator=None):
        """
        @param serializer: A function that takes a single rich input and
            returns a serialized value that can be written out as JSON. May
            raise L{ValidationError} to indicate bad inputs.

        @param extraValidator: Allow additional validation of the field
            value. A callable that takes a field value, and raises
            L{ValidationError} if the value is a incorrect one for this
            field. Alternatively can be set to C{None}, in which case no
            additional validation is done.
        """
        self.key = key
        self.description = description
        self._serializer = serializer
        self._extraValidator = extraValidator

    def validate(self, input):
        """
        Validate the given input value against this L{Field} definition.

        @param input: An input value supposedly serializable by this L{Field}.

        @raises ValidationError: If the value is not serializable or fails to
            be validated by the additional validator.
        """
        # Make sure the input serializes:
        self._serializer(input)
        # Use extra validator, if given:
        if self._extraValidator is not None:
            self._extraValidator(input)

    def serialize(self, input):
        """
        Convert the given input to a value that can actually be logged.

        @param input: An input value supposedly serializable by this L{Field}.

        @return: A serialized value.
        """
        return self._serializer(input)

    @classmethod
    def forValue(klass, key, value, description):
        """
        Create a L{Field} that can only have a single value.

        @param key: The name of the field, the key which refers to it,
            e.g. C{"path"}.

        @param value: The allowed value for the field.

        @param description: A description of what this field contains.
        @type description: C{unicode}

        @return: A L{Field}.
        """

        def validate(checked):
            if checked != value:
                raise ValidationError(checked, "Field %r must be %r" % (key, value))

        return klass(key, lambda _: value, description, validate)

    # PEP 8 variant:
    for_value = forValue

    @classmethod
    def forTypes(klass, key, classes, description, extraValidator=None):
        """
        Create a L{Field} that must be an instance of a given set of types.

        @param key: The name of the field, the key which refers to it,
            e.g. C{"path"}.

        @ivar classes: A C{list} of allowed Python classes for this field's
            values. Supported classes are C{unicode}, C{int}, C{float},
            C{bool}, C{long}, C{list} and C{dict} and C{None} (the latter
            isn't strictly a class, but will be converted appropriately).

        @param description: A description of what this field contains.
        @type description: C{unicode}

        @param extraValidator: See description in L{Field.__init__}.

        @return: A L{Field}.
        """
        fixedClasses = []
        for k in classes:
            if k is None:
                k = type(None)
            if k not in _JSON_TYPES:
                raise TypeError("%s is not JSON-encodeable" % (k,))
            fixedClasses.append(k)
        fixedClasses = tuple(fixedClasses)

        def validate(value):
            if not isinstance(value, fixedClasses):
                raise ValidationError(
                    value, "Field %r requires type to be one of %s" % (key, classes)
                )
            if extraValidator is not None:
                extraValidator(value)

        return klass(key, lambda v: v, description, extraValidator=validate)

    # PEP 8 variant:
    for_types = forTypes


def fields(*fields, **keys):
    """
    Factory for for L{MessageType} and L{ActionType} field definitions.

    @param *fields: A L{tuple} of L{Field} instances.

    @param **keys: A L{dict} mapping key names to the expected type of the
        field's values.

    @return: A L{list} of L{Field} instances.
    """
    return list(fields) + [
        Field.forTypes(key, [value], "") for key, value in keys.items()
    ]


REASON = Field.forTypes(REASON_FIELD, [unicode], "The reason for an event.")
TRACEBACK = Field.forTypes("traceback", [unicode], "The traceback for an exception.")
EXCEPTION = Field.forTypes("exception", [unicode], "The FQPN of an exception class.")


class _MessageSerializer(object):
    """
    A serializer and validator for messages.

    @ivar fields: A C{dict} mapping a C{unicode} field name to the respective
        L{Field}.
    @ivar allow_additional_fields: If true, additional fields don't cause
        validation failure.
    """

    def __init__(self, fields, allow_additional_fields=False):
        keys = []
        for field in fields:
            if not isinstance(field, Field):
                raise TypeError("Expected a Field instance but got", field)
            keys.append(field.key)
        if len(set(keys)) != len(keys):
            raise ValueError(keys, "Duplicate field name")
        if ACTION_TYPE_FIELD in keys:
            if MESSAGE_TYPE_FIELD in keys:
                raise ValueError(
                    keys,
                    "Messages must have either "
                    "'action_type' or 'message_type', not both",
                )
        elif MESSAGE_TYPE_FIELD not in keys:
            raise ValueError(
                keys, "Messages must have either 'action_type' ", "or 'message_type'"
            )
        if any(key.startswith("_") for key in keys):
            raise ValueError(keys, "Field names must not start with '_'")
        for reserved in RESERVED_FIELDS:
            if reserved in keys:
                raise ValueError(
                    keys,
                    "The field name %r is reserved for use "
                    "by the logging framework" % (reserved,),
                )
        self.fields = dict((field.key, field) for field in fields)
        self.allow_additional_fields = allow_additional_fields

    def serialize(self, message):
        """
        Serialize the given message in-place, converting inputs to outputs.

        We do this in-place for performance reasons. There are more fields in
        a message than there are L{Field} objects because of the timestamp,
        task_level and task_uuid fields. By only iterating over our L{Fields}
        we therefore reduce the number of function calls in a critical code
        path.

        @param message: A C{dict}.
        """
        for key, field in self.fields.items():
            message[key] = field.serialize(message[key])

    def validate(self, message):
        """
        Validate the given message.

        @param message: A C{dict}.

        @raises ValidationError: If the message has the wrong fields or one of
            its field values fail validation.
        """
        for key, field in self.fields.items():
            if key not in message:
                raise ValidationError(message, "Field %r is missing" % (key,))
            field.validate(message[key])

        if self.allow_additional_fields:
            return
        # Otherwise, additional fields are not allowed:
        fieldSet = set(self.fields) | set(RESERVED_FIELDS)
        for key in message:
            if key not in fieldSet:
                raise ValidationError(message, "Unexpected field %r" % (key,))


class MessageType(object):
    """
    A specific type of non-action message.

    Example usage:

        # Schema definition:
        KEY = Field("key", [int], u"The lookup key for things.")
        STATUS = Field("status", [int], u"The status of a thing.")
        LOG_STATUS = MessageType(
            "yourapp:subsystem:status", [KEY, STATUS],
            u"We just set the status of something.")

        # Actual code, with logging added:
        def setstatus(key, status):
            doactualset(key, status)
            LOG_STATUS(key=key, status=status).write()

    You do not need to use the L{MessageType} to create the L{eliot.Message},
    however; you could build it up using a series of L{eliot.Message.bind}
    calls. Having a L{MessageType} is nonetheless still useful for validation
    and documentation.

    @ivar message_type: The name of the type,
        e.g. C{"yourapp:subsystem:yourtype"}.

    @ivar description: A description of what this message means.
    @type description: C{unicode}
    """

    def __init__(self, message_type, fields, description=""):
        """
        @ivar type: The name of the type,
            e.g. C{"yourapp:subsystem:yourtype"}.

        @ivar fields: A C{list} of L{Field} instances which can appear in this
            type.

        @param description: A description of what this message means.
        @type description: C{unicode}
        """
        self.message_type = message_type
        self.description = description
        self._serializer = _MessageSerializer(
            fields
            + [Field.forValue(MESSAGE_TYPE_FIELD, message_type, "The message type.")]
        )

    def __call__(self, **fields):
        """
        Create a new L{eliot.Message} of this type with the given fields.

        @param fields: Extra fields to add to the message.

        @rtype: L{eliot.Message}
        """
        warn(
            "MessageType.__call__() is deprecated since 1.11.0, "
            "use MessageType.log() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        fields[MESSAGE_TYPE_FIELD] = self.message_type
        return Message(fields, self._serializer)

    def log(self, **fields):
        """
        Write a new L{Message} of this type to the default L{Logger}.

        The keyword arguments will become contents of the L{Message}.
        """
        fields["__eliot_serializer__"] = self._serializer
        log_message(self.message_type, **fields)


class _ActionSerializers(PClass):
    """
    Serializers for the three action messages: start, success and failure.
    """

    start = pyrsistent_field(mandatory=True)
    success = pyrsistent_field(mandatory=True)
    failure = pyrsistent_field(mandatory=True)


class ActionType(object):
    """
    A specific type of action.

    Example usage:

        # Schema definition:
        KEY = Field("key", [int], u"The lookup key for things.")
        RESULT = Field("result", [str], u"The result of lookups.")
        LOG_DOSOMETHING = ActionType(
            "yourapp:subsystem:youraction",
            [KEY], [RESULT],
            u"Do something with a key, resulting in a value.")

        # Actual code, with logging added:
        def dosomething(key):
            with LOG_DOSOMETHING(logger, key=key) as action:
                _dostuff(key)
                _morestuff(key)
                result = _theresult()
                action.addSuccessFields(result=result)
            return result

    @ivar action_type: The name of the action,
        e.g. C{"yourapp:subsystem:youraction"}.

    @ivar startFields: A C{list} of L{Field} instances which can appear in
        this action's start message.

    @ivar successFields: A C{list} of L{Field} instances which can appear in
        this action's successful finish message.

    @ivar failureFields: A C{list} of L{Field} instances which can appear in
        this action's failed finish message (in addition to the built-in
        C{"exception"} and C{"reason"} fields).

    @ivar description: A description of what this action's messages mean.
    @type description: C{unicode}
    """

    # Overrideable hook for testing; need staticmethod() so functions don't
    # get turned into methods.
    _start_action = staticmethod(start_action)
    _startTask = staticmethod(startTask)

    def __init__(self, action_type, startFields, successFields, description=""):
        self.action_type = action_type
        self.description = description

        actionTypeField = Field.forValue(
            ACTION_TYPE_FIELD, action_type, "The action type"
        )

        def makeActionStatusField(value):
            return Field.forValue(ACTION_STATUS_FIELD, value, "The action status")

        startFields = startFields + [
            actionTypeField,
            makeActionStatusField(STARTED_STATUS),
        ]
        successFields = successFields + [
            actionTypeField,
            makeActionStatusField(SUCCEEDED_STATUS),
        ]
        failureFields = [
            actionTypeField,
            makeActionStatusField(FAILED_STATUS),
            REASON,
            EXCEPTION,
        ]

        self._serializers = _ActionSerializers(
            start=_MessageSerializer(startFields),
            success=_MessageSerializer(successFields),
            # Failed action messages can have extra fields from exception
            # extraction:
            failure=_MessageSerializer(failureFields, allow_additional_fields=True),
        )

    def __call__(self, logger=None, **fields):
        """
        Start a new L{eliot.Action} of this type with the given start fields.

        You can use the result as a Python context manager, or use the
        L{eliot.Action.finish} API.

             LOG_DOSOMETHING = ActionType("yourapp:subsystem:dosomething",
                                      [Field.forTypes("entry", [int], "")],
                                      [Field.forTypes("result", [int], "")],
                                      [],
                                      "Do something with an entry.")
             with LOG_DOSOMETHING(entry=x) as action:
                  do(x)
                  result = something(x * 2)
                  action.addSuccessFields(result=result)

        Or perhaps:

             action = LOG_DOSOMETHING(entry=x)
             action.run(doSomething)
             action.finish()

        @param logger: A L{eliot.ILogger} provider to which the action's
            messages will be written, or C{None} to use the default one.

        @param fields: Extra fields to add to the message.

        @rtype: L{eliot.Action}
        """
        return self._start_action(logger, self.action_type, self._serializers, **fields)

    def as_task(self, logger=None, **fields):
        """
        Start a new L{eliot.Action} of this type as a task (i.e. top-level
        action) with the given start fields.

        See L{ActionType.__call__} for example of usage.

        @param logger: A L{eliot.ILogger} provider to which the action's
            messages will be written, or C{None} to use the default one.

        @param fields: Extra fields to add to the message.

        @rtype: L{eliot.Action}
        """
        return self._startTask(logger, self.action_type, self._serializers, **fields)

    # Backwards compatible variant:
    asTask = as_task


__all__ = []
